"""This module contains an SMT abstraction of arrays.

This includes an Array class to implement basic store and set
operations, as well as as a K-array, which can be initialized with
default values over a certain range.
"""

from typing import cast
import z3

from mythril.laser.smt.bitvec import BitVec


class BaseArray:
    """Base array type, which implements basic store and set operations."""

    def __init__(self, raw):
        self.raw = raw

    def __getitem__(self, item: BitVec) -> BitVec:
        """Gets item from the array, item can be symbolic."""
        if isinstance(item, slice):
            raise ValueError(
                "Instance of BaseArray, does not support getitem with slices"
            )
        return BitVec(cast(z3.BitVecRef, z3.Select(self.raw, item.raw)))

    def __setitem__(self, key: BitVec, value: BitVec) -> None:
        """Sets an item in the array, key can be symbolic."""
        self.raw = z3.Store(self.raw, key.raw, value.raw)

    def substitute(self, original_expression, new_expression):
        """

        :param original_expression:
        :param new_expression:
        """
        if self.raw is None:
            return
        original_z3 = original_expression.raw
        new_z3 = new_expression.raw
        self.raw = z3.substitute(self.raw, (original_z3, new_z3))


class Array(BaseArray):
    """A basic symbolic array."""

    def __init__(self, name: str, domain: int, value_range: int):
        """Initializes a symbolic array.

        :param name: Name of the array
        :param domain: The domain for the array (10 -> all the values that a bv of size 10 could take)
        :param value_range: The range for the values in the array (10 -> all the values that a bv of size 10 could take)
        """
        self.domain = z3.BitVecSort(domain)
        self.range = z3.BitVecSort(value_range)
        super(Array, self).__init__(z3.Array(name, self.domain, self.range))


class K(BaseArray):
    """A basic symbolic array, which can be initialized with a default
    value."""

    def __init__(self, domain: int, value_range: int, value: int):
        """Initializes an array with a default value.

        :param domain: The domain for the array (10 -> all the values that a bv of size 10 could take)
        :param value_range: The range for the values in the array (10 -> all the values that a bv of size 10 could take)
        :param value: The default value to use for this array
        """
        self.domain = z3.BitVecSort(domain)
        self.value = z3.BitVecVal(value, value_range)
        self.raw = z3.K(self.domain, self.value)



"""This module contains various utility functions regarding unit conversion and
solc integration."""
import binascii
import json
import sys
import os
import platform
import logging
import solc
import re
import typing

from pathlib import Path
from requests.exceptions import ConnectionError
from subprocess import PIPE, Popen


from json.decoder import JSONDecodeError
import semantic_version as semver
from semantic_version import Version, NpmSpec
from pyparsing import Word, Group, Optional, ZeroOrMore, oneOf, Regex, Combine

from mythril.exceptions import CompilerError
from mythril.support.support_args import args

import solcx

log = logging.getLogger(__name__)


def safe_decode(hex_encoded_string):
    """

    :param hex_encoded_string:
    :return:
    """
    if hex_encoded_string.startswith("0x"):
        return bytes.fromhex(hex_encoded_string[2:])
    else:
        return bytes.fromhex(hex_encoded_string)


def get_solc_json(file, solc_binary="solc", solc_settings_json=None):
    """

    :param file:
    :param solc_binary:
    :param solc_settings_json:
    :return:
    """
    if args.solc_args is None:
        cmd = [solc_binary, "--standard-json", "--allow-paths", ".,/"]
    else:
        cmd = [solc_binary, "--standard-json"] + args.solc_args.split()

    settings = {}
    if solc_settings_json:
        with open(solc_settings_json) as f:
            settings = json.load(f)
    if "optimizer" not in settings:
        settings.update({"optimizer": {"enabled": False}})

    settings.update(
        {
            "outputSelection": {
                "*": {
                    "": ["ast"],
                    "*": [
                        "metadata",
                        "evm.bytecode",
                        "evm.deployedBytecode",
                        "evm.methodIdentifiers",
                    ],
                }
            },
        }
    )

    input_json = json.dumps(
        {
            "language": "Solidity",
            "sources": {file: {"urls": [file]}},
            "settings": settings,
        }
    )

    try:
        p = Popen(cmd, stdin=PIPE, stdout=PIPE, stderr=PIPE)
        stdout, stderr = p.communicate(bytes(input_json, "utf8"))

    except FileNotFoundError:
        raise CompilerError(
            "Compiler not found. Make sure that solc is installed and in PATH, or set the SOLC environment variable."
        )

    out = stdout.decode("UTF-8")

    try:
        result = json.loads(out)
    except JSONDecodeError as e:
        log.error(f"Encountered a decode error.\n stdout:{out}\n stderr: {stderr}")
        raise e

    for error in result.get("errors", []):
        if error["severity"] == "error":
            raise CompilerError(
                "Solc experienced a fatal error.\n\n%s" % error["formattedMessage"]
            )

    return result


def get_random_address():
    """

    :return:
    """
    return binascii.b2a_hex(os.urandom(20)).decode("UTF-8")


def get_indexed_address(index):
    """

    :param index:
    :return:
    """
    return "0x" + (hex(index)[2:] * 40)


def solc_exists(version):
    """

    :param version:
    :return:
    """

    default_binary = "/usr/bin/solc"
    if platform.system() == "Darwin":
        solcx.import_installed_solc()
    solcx.install_solc("v" + version)
    solcx.set_solc_version("v" + version)
    solc_binary = solcx.install.get_executable()
    return solc_binary


def parse_pragma(solidity_code):
    lt = Word("<")
    gtr = Word(">")
    eq = Word("=")
    carrot = Word("^")
    version = Regex(r"\s*[0-9]+\s*\.\s*[0-9]+\s*(\.\s*[0-9]+)?")
    inequality = Optional(
        eq | (Combine(gtr + Optional(eq)) | Combine(lt + Optional(eq)))
    )
    min_version = Optional(carrot | inequality) + version
    max_version = Optional(inequality + version)
    pragma = Word("pragma") + Word("solidity") + min_version + Optional(max_version)
    result = pragma.parseString(solidity_code)
    min_inequality = result[2] if result[2] in [">", "<", ">=", "<="] else ""
    min_carrot = result[2] if result[2] == "^" else ""
    min_version = result[3] if min_carrot != "" or min_inequality != "" else result[2]
    return {
        "min_carrot": min_carrot,
        "min_inequality": min_inequality,
        "min_version": min_version,
        "max_inequality": result[4] if len(result) > 4 else None,
        "max_version": result[5] if len(result) > 5 else None,
    }


try:
    all_versions = solcx.get_installable_solc_versions()
except ConnectionError:
    # No internet, trying to proceed with installed compilers
    all_versions = solcx.get_installed_solc_versions()


def extract_version(file: typing.Optional[str]):
    if file is None:
        return None
    version_line = None
    for line in file.split("\n"):
        if "pragma solidity" not in line:
            continue
        version_line = line.rstrip()
        break
    if version_line is None:
        return None

    assert "pragma solidity" in version_line
    if version_line[-1] == ";":
        version_line = version_line[:-1]
    version_line = version_line[version_line.find("pragma") :]
    pragma_dict = parse_pragma(version_line)

    min_inequality = pragma_dict.get("min_inequality", None)
    max_inequality = pragma_dict.get("max_inequality", None)
    min_version = pragma_dict.get("min_version", None)
    if min_version is not None:
        min_version = min_version.replace(" ", "").replace("\t", "")
    max_version = pragma_dict.get("max_version", None)
    if max_version is not None:
        max_version = max_version.replace(" ", "").replace("\t", "")

    version_spec = (
        f"{min_inequality}{min_version},{max_inequality}{max_version}"
        if max_version
        else min_version
    )
    version_constraint = semver.SimpleSpec(version_spec)

    for version in all_versions:
        if version in version_constraint:
            if "0.5.17" in str(version):
                # Solidity 0.5.17 Does not compile in a lot of cases.
                continue
            return str(version)

    else:
        return None

    return None


def extract_binary(file: str) -> str:
    file_data = None
    with open(file) as f:
        file_data = f.read()

    version = extract_version(file_data)
    if (
        version
        and NpmSpec("^0.8.0").match(Version(version))
        and "unchecked" not in file_data
    ):
        args.use_integer_module = False

    if version is None:
        return os.environ.get("SOLC") or "solc"
    return solc_exists(version)



from mythril.laser.smt import (
    BitVec,
    Function,
    URem,
    symbol_factory,
    ULE,
    And,
    ULT,
    Bool,
    Or,
)

from mythril.support.support_utils import sha3
from typing import Dict, Tuple, List, Optional

import logging


TOTAL_PARTS = 10**40
PART = (2**256 - 1) // TOTAL_PARTS
INTERVAL_DIFFERENCE = 10**30
log = logging.getLogger(__name__)


class KeccakFunctionManager:
    """
    A bunch of uninterpreted functions are considered like keccak256_160 ,...
    where keccak256_160 means the input of keccak256() is 160 bit number.
    the range of these functions are constrained to some mutually disjoint intervals
    All the hashes modulo 64 are 0 as we need a spread among hashes for array type data structures
    All the functions are kind of one to one due to constraint of the existence of inverse
    for each encountered input.
    For more info https://files.sri.inf.ethz.ch/website/papers/sp20-verx.pdf
    """

    hash_matcher = "fffffff"  # This is usually the prefix for the hash in the output

    def __init__(self):
        self.store_function: Dict[int, Tuple[Function, Function]] = {}
        self.interval_hook_for_size: Dict[int, int] = {}
        self._index_counter = TOTAL_PARTS - 34534
        self.hash_result_store: Dict[int, List[BitVec]] = {}

        self.quick_inverse: Dict[BitVec, BitVec] = {}  # This is for VMTests
        self.concrete_hashes: Dict[BitVec, BitVec] = {}
        self.symbolic_inputs: Dict[int, List[BitVec]] = {}

    def reset(self):
        self.store_function = {}
        self.interval_hook_for_size = {}
        self.hash_result_store: Dict[int, List[BitVec]] = {}
        self.quick_inverse = {}
        self.concrete_hashes = {}
        self.symbolic_inputs = {}

    @staticmethod
    def find_concrete_keccak(data: BitVec) -> BitVec:
        """
        Calculates concrete keccak
        :param data: input bitvecval
        :return: concrete keccak output
        """
        keccak = symbol_factory.BitVecVal(
            int.from_bytes(
                sha3(data.value.to_bytes(data.size() // 8, byteorder="big")), "big"
            ),
            256,
        )
        return keccak

    def get_function(self, length: int) -> Tuple[Function, Function]:
        """
        Returns the keccak functions for the corresponding length
        :param length: input size
        :return: tuple of keccak and it's inverse
        """
        try:
            func, inverse = self.store_function[length]
        except KeyError:
            func = Function("keccak256_{}".format(length), [length], 256)
            inverse = Function("keccak256_{}-1".format(length), [256], length)
            self.store_function[length] = (func, inverse)
            self.hash_result_store[length] = []
        return func, inverse

    @staticmethod
    def get_empty_keccak_hash() -> BitVec:
        """
        returns sha3("")
        :return:
        """
        val = 89477152217924674838424037953991966239322087453347756267410168184682657981552
        return symbol_factory.BitVecVal(val, 256)

    def create_keccak(self, data: BitVec) -> BitVec:
        """
        Creates Keccak of the data
        :param data: input
        :return: Tuple of keccak and the condition it should satisfy
        """
        length = data.size()
        func, _ = self.get_function(length)

        if data.symbolic is False:
            concrete_hash = self.find_concrete_keccak(data)
            self.concrete_hashes[data] = concrete_hash
            return concrete_hash

        if length not in self.symbolic_inputs:
            self.symbolic_inputs[length] = []

        self.symbolic_inputs[length].append(data)
        self.hash_result_store[length].append(func(data))
        return func(data)

    def create_conditions(self) -> Bool:
        condition = symbol_factory.Bool(True)
        for inputs_list in self.symbolic_inputs.values():
            for symbolic_input in inputs_list:
                condition = And(
                    condition, self._create_condition(func_input=symbolic_input)
                )
        for concrete_input, concrete_hash in self.concrete_hashes.items():
            func, inverse = self.get_function(concrete_input.size())
            condition = And(
                condition,
                func(concrete_input) == concrete_hash,
                inverse(func(concrete_input)) == concrete_input,
            )
        return condition

    def get_concrete_hash_data(self, model) -> Dict[int, List[Optional[int]]]:
        """
        returns concrete values of hashes in the self.hash_result_store
        :param model: The z3 model to query for concrete values
        :return: A dictionary with concrete hashes { <hash_input_size> : [<concrete_hash>, <concrete_hash>]}
        """
        concrete_hashes = {}  # type: Dict[int, List[Optional[int]]]
        for size in self.hash_result_store:
            concrete_hashes[size] = []
            for val in self.hash_result_store[size]:
                eval_ = model.eval(val.raw)
                try:
                    concrete_val = eval_.as_long()
                    concrete_hashes[size].append(concrete_val)
                except AttributeError:
                    continue
        return concrete_hashes

    def _create_condition(self, func_input: BitVec) -> Bool:
        """
        Creates the constraints for hash
        :param func_input: input of the hash
        :return: condition
        """
        length = func_input.size()
        func, inv = self.get_function(length)
        try:
            index = self.interval_hook_for_size[length]
        except KeyError:
            self.interval_hook_for_size[length] = self._index_counter
            index = self._index_counter
            self._index_counter -= INTERVAL_DIFFERENCE

        lower_bound = index * PART
        upper_bound = lower_bound + PART

        cond = And(
            inv(func(func_input)) == func_input,
            ULE(symbol_factory.BitVecVal(lower_bound, 256), func(func_input)),
            ULT(func(func_input), symbol_factory.BitVecVal(upper_bound, 256)),
            URem(func(func_input), symbol_factory.BitVecVal(64, 256)) == 0,
        )
        concrete_cond = symbol_factory.Bool(False)
        for key, keccak in self.concrete_hashes.items():
            if key.size() == func_input.size():
                hash_eq = And(func(func_input) == keccak, key == func_input)
                concrete_cond = Or(concrete_cond, hash_eq)
        return And(inv(func(func_input)) == func_input, Or(cond, concrete_cond))


keccak_function_manager = KeccakFunctionManager()



"""This module contains an abstract SMT representation of an SMT solver."""
import logging
import os
import sys
import z3
from typing import Union, cast, TypeVar, Generic, List, Sequence

from mythril.laser.smt.expression import Expression
from mythril.laser.smt.model import Model
from mythril.laser.smt.bool import Bool
from mythril.laser.smt.solver.solver_statistics import stat_smt_query

T = TypeVar("T", bound=Union[z3.Solver, z3.Optimize])

log = logging.getLogger(__name__)


class BaseSolver(Generic[T]):
    def __init__(self, raw: T) -> None:
        """"""
        self.raw = raw

    def set_timeout(self, timeout: int) -> None:
        """Sets the timeout that will be used by this solver, timeout is in
        milliseconds.

        :param timeout:
        """
        self.raw.set(timeout=timeout)

    def add(self, *constraints: List[Bool]) -> None:
        """Adds the constraints to this solver.

        :param constraints:
        :return:
        """
        z3_constraints = [
            c.raw for c in cast(List[Bool], constraints)
        ]  # type: Sequence[z3.BoolRef]
        self.raw.add(z3_constraints)

    def append(self, *constraints: List[Bool]) -> None:
        """Adds the constraints to this solver.

        :param constraints:
        :return:
        """
        self.add(*constraints)

    @stat_smt_query
    def check(self, *args) -> z3.CheckSatResult:
        """Returns z3 smt check result.
        Also suppresses the stdout when running z3 library's check() to avoid unnecessary output
        :return: The evaluated result which is either of sat, unsat or unknown
        """
        old_stdout = sys.stdout
        with open(os.devnull, "w") as dev_null_fd:
            sys.stdout = dev_null_fd
            try:
                evaluate = self.raw.check(args)
            except z3.z3types.Z3Exception as e:
                # Some requests crash the solver
                evaluate = z3.unknown
                log.info(f"Encountered Z3 exception when checking the constraints: {e}")
        sys.stdout = old_stdout
        return evaluate

    def model(self) -> Model:
        """Returns z3 model for a solution.

        :return:
        """
        try:
            return Model([self.raw.model()])
        except z3.z3types.Z3Exception as e:
            log.info(f"Encountered a Z3 exception while querying for the model: {e}")
            return Model()

    def sexpr(self):
        return self.raw.sexpr()


class Solver(BaseSolver[z3.Solver]):
    """An SMT solver object."""

    def __init__(self) -> None:
        """"""
        super().__init__(z3.Solver())

    def reset(self) -> None:
        """Reset this solver."""
        self.raw.reset()

    def pop(self, num: int) -> None:
        """Pop num constraints from this solver.

        :param num:
        """
        self.raw.pop(num)


class Optimize(BaseSolver[z3.Optimize]):
    """An optimizing smt solver."""

    def __init__(self) -> None:
        """Create a new optimizing solver instance."""
        super().__init__(z3.Optimize())

    def minimize(self, element: Expression[z3.ExprRef]) -> None:
        """In solving this solver will try to minimize the passed expression.

        :param element:
        """
        self.raw.minimize(element.raw)

    def maximize(self, element: Expression[z3.ExprRef]) -> None:
        """In solving this solver will try to maximize the passed expression.

        :param element:
        """
        self.raw.maximize(element.raw)



"""This module includes classes used for annotating trace information.

This includes the base StateAnnotation class, as well as an adaption,
which will not be copied on every new state.
"""

from abc import abstractmethod


class StateAnnotation:
    """The StateAnnotation class is used to persist information over traces.

    This allows modules to reason about traces without the need to
    traverse the state space themselves.
    """

    # TODO: Remove this? It seems to be used only in the MutationPruner, and
    # we could simply use world state annotations if we want them to be persisted.
    @property
    def persist_to_world_state(self) -> bool:
        """If this function returns true then laser will also annotate the
        world state.

        If you want annotations to persist through different user initiated message call transactions
        then this should be enabled.

        The default is set to False
        """
        return False

    @property
    def persist_over_calls(self) -> bool:
        """If this function returns true then laser will propagate the annotation between calls

        The default is set to False
        """
        return False

    @property
    def search_importance(self) -> int:
        """
        Used in estimating the priority of a state annotated with the corresponding annotation.
        Default is 1
        """
        return 1


class MergeableStateAnnotation(StateAnnotation):
    """This class allows a base annotation class for annotations that
    can be merged.
    """

    @abstractmethod
    def check_merge_annotation(self, annotation) -> bool:
        pass

    @abstractmethod
    def merge_annotation(self, annotation):
        pass


class NoCopyAnnotation(StateAnnotation):
    """This class provides a base annotation class for annotations that
    shouldn't be copied on every new state.

    Rather the same object should be propagated. This is very useful if
    you are looking to analyze a property over multiple substates
    """

    def __copy__(self):
        return self

    def __deepcopy__(self, _):
        return self



from typing import List

from mythril.laser.ethereum.state.global_state import GlobalState
from . import BasicSearchStrategy


class BeamSearch(BasicSearchStrategy):
    """chooses a random state from the worklist with equal likelihood."""

    def __init__(self, work_list, max_depth, beam_width, **kwargs):
        super().__init__(work_list, max_depth)
        self.beam_width = beam_width

    @staticmethod
    def beam_priority(state):
        return sum([annotation.search_importance for annotation in state._annotations])

    def sort_and_eliminate_states(self):
        self.work_list.sort(key=lambda state: self.beam_priority(state), reverse=True)
        del self.work_list[self.beam_width :]

    def view_strategic_global_state(self) -> GlobalState:
        """

        :return:
        """
        self.sort_and_eliminate_states()
        if len(self.work_list) > 0:
            return self.work_list[0]
        else:
            raise IndexError

    def get_strategic_global_state(self) -> GlobalState:
        """

        :return:
        """
        self.sort_and_eliminate_states()
        if len(self.work_list) > 0:
            return self.work_list.pop(0)
        else:
            raise IndexError



"""This module contains the detection code for predictable variable
dependence."""
import logging
from copy import copy
from mythril.analysis.issue_annotation import IssueAnnotation
from mythril.analysis.module.base import DetectionModule, EntryPoint
from mythril.analysis.report import Issue
from mythril.exceptions import UnsatError
from mythril.analysis import solver
from mythril.analysis.swc_data import TX_ORIGIN_USAGE
from mythril.laser.ethereum.state.global_state import GlobalState
from mythril.laser.smt import And
from typing import List

log = logging.getLogger(__name__)


class TxOriginAnnotation:
    """Symbol annotation added to a variable that is initialized with a call to the ORIGIN instruction."""

    def __init__(self) -> None:
        pass


class TxOrigin(DetectionModule):
    """This module detects whether control flow decisions are made based on the transaction origin."""

    name = "Control flow depends on tx.origin"
    swc_id = TX_ORIGIN_USAGE
    description = "Check whether control flow decisions are influenced by tx.origin"
    entry_point = EntryPoint.CALLBACK

    pre_hooks = ["JUMPI"]
    post_hooks = ["ORIGIN"]

    def _execute(self, state: GlobalState) -> List[Issue]:
        """

        :param state:
        :return:
        """
        return self._analyze_state(state)

    def _analyze_state(self, state: GlobalState) -> List[Issue]:
        """

        :param state:
        :return:
        """

        issues = []

        if state.get_current_instruction()["opcode"] == "JUMPI":
            # We're in JUMPI prehook

            for annotation in state.mstate.stack[-2].annotations:

                if isinstance(annotation, TxOriginAnnotation):
                    constraints = copy(state.world_state.constraints)

                    try:
                        transaction_sequence = solver.get_transaction_sequence(
                            state, constraints
                        )
                    except UnsatError:
                        continue

                    description = (
                        "The tx.origin environment variable has been found to influence a control flow decision. "
                        "Note that using tx.origin as a security control might cause a situation where a user "
                        "inadvertently authorizes a smart contract to perform an action on their behalf. It is "
                        "recommended to use msg.sender instead."
                    )

                    severity = "Low"

                    """
                    Note: We report the location of the JUMPI instruction. Usually this maps to an if or
                    require statement.
                    """

                    issue = Issue(
                        contract=state.environment.active_account.contract_name,
                        function_name=state.environment.active_function_name,
                        address=state.get_current_instruction()["address"],
                        swc_id=TX_ORIGIN_USAGE,
                        bytecode=state.environment.code.bytecode,
                        title="Dependence on tx.origin",
                        severity=severity,
                        description_head="Use of tx.origin as a part of authorization control.",
                        description_tail=description,
                        gas_used=(state.mstate.min_gas_used, state.mstate.max_gas_used),
                        transaction_sequence=transaction_sequence,
                    )
                    state.annotate(
                        IssueAnnotation(
                            conditions=[And(*constraints)], issue=issue, detector=self
                        )
                    )

                    issues.append(issue)

        else:

            # In ORIGIN posthook

            state.mstate.stack[-1].annotate(TxOriginAnnotation())

        return issues


detector = TxOrigin()



"""This module contains general exceptions used by Mythril."""


class MythrilBaseException(Exception):
    """The Mythril exception base type."""

    pass


class CompilerError(MythrilBaseException):
    """A Mythril exception denoting an error during code compilation."""

    pass


class UnsatError(MythrilBaseException):
    """A Mythril exception denoting the unsatisfiability of a series of
    constraints."""

    pass


class SolverTimeOutException(UnsatError):
    """A Mythril exception denoting the unsatisfiability of a series of
    constraints."""

    pass


class NoContractFoundError(MythrilBaseException):
    """A Mythril exception denoting that a given contract file was not
    found."""

    pass


class CriticalError(MythrilBaseException):
    """A Mythril exception denoting an unknown critical error has been
    encountered."""

    pass


class DetectorNotFoundError(MythrilBaseException):
    """A Mythril exception denoting attempted usage of a non-existant
    detection module."""

    pass


class IllegalArgumentError(ValueError):
    """The argument used does not exist"""

    pass



from mythril.laser.ethereum.strategy import BasicSearchStrategy
from mythril.laser.ethereum.state.global_state import GlobalState
from mythril.laser.plugin.plugins.coverage import InstructionCoveragePlugin


class CoverageStrategy(BasicSearchStrategy):
    """Implements a instruction coverage based search strategy

    This strategy is quite simple and effective, it prioritizes the execution of instructions that have previously been
    uncovered. Once there is no such global state left in the work list, it will resort to using the super_strategy.

    This strategy is intended to be used "on top of" another one
    """

    def __init__(
        self,
        super_strategy: BasicSearchStrategy,
        instruction_coverage_plugin: InstructionCoveragePlugin,
    ):
        self.super_strategy = super_strategy
        self.instruction_coverage_plugin = instruction_coverage_plugin
        BasicSearchStrategy.__init__(
            self, super_strategy.work_list, super_strategy.max_depth
        )

    def get_strategic_global_state(self) -> GlobalState:
        """
        Returns the first uncovered global state in the work list if it exists,
        otherwise super_strategy.get_strategic_global_state() is returned.
        """
        for global_state in self.work_list:
            if not self._is_covered(global_state):
                self.work_list.remove(global_state)
                return global_state
        return self.super_strategy.get_strategic_global_state()

    def _is_covered(self, global_state: GlobalState) -> bool:
        """Checks if the instruction for the given global state is already covered"""
        bytecode = global_state.environment.code.bytecode
        index = global_state.mstate.pc
        return self.instruction_coverage_plugin.is_instruction_covered(bytecode, index)



from mythril.laser.ethereum.state.global_state import GlobalState
from mythril.laser.ethereum.state.constraints import Constraints
from mythril.laser.ethereum.strategy.basic import BasicSearchStrategy
from mythril.laser.ethereum.state.annotation import StateAnnotation
from mythril.laser.ethereum.transaction import ContractCreationTransaction
from mythril.laser.ethereum.util import get_instruction_index
from mythril.analysis.solver import get_transaction_sequence
from mythril.laser.smt import Not
from mythril.exceptions import UnsatError

from functools import reduce
from typing import Dict, cast, List, Any, Tuple
from copy import copy
from . import CriterionSearchStrategy
import logging
import operator

log = logging.getLogger(__name__)


class TraceAnnotation(StateAnnotation):
    """
    This is the annotation used by the ConcolicStrategy to store concolic traces.
    """

    def __init__(self, trace=None):
        self.trace = trace or []

    @property
    def persist_over_calls(self) -> bool:
        return True

    def __copy__(self):
        return TraceAnnotation(copy(self.trace))


class ConcolicStrategy(CriterionSearchStrategy):
    """
    Executes program concolically using the input trace till a specific branch
    """

    def __init__(
        self,
        work_list: List[GlobalState],
        max_depth: int,
        trace: List[List[Tuple[int, str]]],
        flip_branch_addresses: List[str],
    ):
        """

        work_list: The work-list of states
        max_depth: The maximum depth for the strategy to go through
        trace: This is the trace to be followed, each element is of the type Tuple(program counter, tx_id)
        flip_branch_addresses: Branch addresses to be flipped.
        """
        super().__init__(work_list, max_depth)
        self.trace: List[Tuple[int, str]] = reduce(operator.iconcat, trace, [])
        self.last_tx_count: int = len(trace)
        self.flip_branch_addresses: List[str] = flip_branch_addresses
        self.results: Dict[str, Dict[str, Any]] = {}

    def check_completion_criterion(self):
        if len(self.flip_branch_addresses) == len(self.results):
            self.set_criterion_satisfied()

    def get_strategic_global_state(self) -> GlobalState:
        """
        This function does the following:-
        1) Choose the states by following the concolic trace.
        2) In case we have an executed JUMPI that is in flip_branch_addresses, flip that branch.
        :return:
        """
        while len(self.work_list) > 0:
            state = self.work_list.pop()
            seq_id = len(state.world_state.transaction_sequence)

            trace_annotations = cast(
                List[TraceAnnotation],
                list(state.world_state.get_annotations(TraceAnnotation)),
            )

            if len(trace_annotations) == 0:
                annotation = TraceAnnotation()
                state.world_state.annotate(annotation)
            else:
                annotation = trace_annotations[0]

            # Appends trace
            annotation.trace.append((state.mstate.pc, state.current_transaction.id))

            # If length of trace is 1 then it is not a JUMPI
            if len(annotation.trace) < 2:
                # If trace does not follow the specified path, ignore the state
                if annotation.trace != self.trace[: len(annotation.trace)]:
                    continue
                return state

            # Get the address of the previous pc
            addr: str = str(
                state.environment.code.instruction_list[annotation.trace[-2][0]][
                    "address"
                ]
            )
            if (
                annotation.trace == self.trace[: len(annotation.trace)]
                and seq_id == self.last_tx_count
                and addr in self.flip_branch_addresses
                and addr not in self.results
            ):
                if (
                    state.environment.code.instruction_list[annotation.trace[-2][0]][
                        "opcode"
                    ]
                    != "JUMPI"
                ):
                    log.error(
                        f"The branch {addr} does not lead "
                        "to a jump address, skipping this branch"
                    )
                    continue

                constraints = Constraints(state.world_state.constraints[:-1])
                constraints.append(Not(state.world_state.constraints[-1]))

                try:
                    self.results[addr] = get_transaction_sequence(state, constraints)
                except UnsatError:
                    self.results[addr] = None
            elif annotation.trace != self.trace[: len(annotation.trace)]:
                continue
            self.check_completion_criterion()
            return state
        raise StopIteration



from mythril.laser.ethereum.state.global_state import GlobalState
from mythril.laser.ethereum.strategy.basic import BasicSearchStrategy
from mythril.laser.ethereum.state.annotation import StateAnnotation
from mythril.laser.ethereum.transaction import ContractCreationTransaction
from mythril.laser.smt import And, simplify
from mythril.support.support_utils import ModelCache

from typing import Dict, cast, List
from collections import OrderedDict
from copy import copy, deepcopy
from functools import lru_cache
import logging
import z3
from time import time

log = logging.getLogger(__name__)


class DelayConstraintStrategy(BasicSearchStrategy):
    def __init__(self, work_list, max_depth, **kwargs):
        super().__init__(work_list, max_depth)
        self.model_cache = ModelCache()
        self.pending_worklist = []
        log.info("Loaded search strategy extension: DelayConstraintStrategy")

    def get_strategic_global_state(self) -> GlobalState:
        """Returns the next state

        :return: Global state
        """
        while True:
            while len(self.work_list) == 0:
                state = self.pending_worklist.pop(0)
                model = state.world_state.constraints.get_model()
                if model is not None:
                    self.model_cache.put(model, 1)
                    self.work_list.append(state)
            state = self.work_list.pop(0)
            return state



from mythril.analysis.module import DetectionModule

from mythril.plugin.interface import MythrilPlugin, MythrilLaserPlugin
from mythril.plugin.discovery import PluginDiscovery
from mythril.support.support_utils import Singleton

from mythril.analysis.module.loader import ModuleLoader
from mythril.laser.plugin.loader import LaserPluginLoader
from typing import Dict
import logging

log = logging.getLogger(__name__)


class UnsupportedPluginType(Exception):
    """Raised when a plugin with an unsupported type is loaded"""

    pass


class MythrilPluginLoader(object, metaclass=Singleton):
    """MythrilPluginLoader singleton

    This object permits loading MythrilPlugin's
    """

    def __init__(self):
        log.info("Initializing mythril plugin loader")
        self.loaded_plugins = []
        self.plugin_args = dict()  # type: Dict[str, Dict]
        self._load_default_enabled()

    def set_args(self, plugin_name: str, **kwargs):
        self.plugin_args[plugin_name] = kwargs

    def load(self, plugin: MythrilPlugin):
        """Loads the passed plugin

        This function handles input validation and dispatches loading to type specific loaders.
        Supported plugin types:
         - laser plugins
         - detection modules
        """
        if not isinstance(plugin, MythrilPlugin):
            raise ValueError("Passed plugin is not of type MythrilPlugin")
        logging.info(f"Loading plugin: {plugin.name}")

        log.info(f"Loading plugin: {str(plugin)}")
        if isinstance(plugin, DetectionModule):
            self._load_detection_module(plugin)
        elif isinstance(plugin, MythrilLaserPlugin):
            self._load_laser_plugin(plugin)
        else:
            raise UnsupportedPluginType("Passed plugin type is not yet supported")

        self.loaded_plugins.append(plugin)
        log.info(f"Finished loading plugin: {plugin.name}")

    @staticmethod
    def _load_detection_module(plugin: DetectionModule):
        """Loads the passed detection module"""
        log.info(f"Loading detection module: {plugin.name}")
        ModuleLoader().register_module(plugin)

    @staticmethod
    def _load_laser_plugin(plugin: MythrilLaserPlugin):
        """Loads the laser plugin"""
        log.info(f"Loading laser plugin: {plugin.name}")
        LaserPluginLoader().load(plugin)

    def _load_default_enabled(self):
        """Loads the plugins that have the default enabled flag"""
        log.info("Loading installed analysis modules that are enabled by default")
        for plugin_name in PluginDiscovery().get_plugins(default_enabled=True):
            plugin = PluginDiscovery().build_plugin(
                plugin_name, self.plugin_args.get(plugin_name, {})
            )
            self.load(plugin)



from mythril.laser.smt import Solver, symbol_factory
import z3


def test_decls():
    # Arrange
    solver = Solver()
    x = symbol_factory.BitVecSym("x", 256)
    expression = x == symbol_factory.BitVecVal(2, 256)

    # Act
    solver.add(expression)
    result = solver.check()
    model = solver.model()

    decls = model.decls()

    # Assert
    assert z3.sat == result
    assert x.raw.decl() in decls


def test_get_item():
    # Arrange
    solver = Solver()
    x = symbol_factory.BitVecSym("x", 256)
    expression = x == symbol_factory.BitVecVal(2, 256)

    # Act
    solver.add(expression)
    result = solver.check()
    model = solver.model()

    x_concrete = model[x.raw.decl()]

    # Assert
    assert z3.sat == result
    assert 2 == x_concrete


def test_as_long():
    # Arrange
    solver = Solver()
    x = symbol_factory.BitVecSym("x", 256)
    expression = x == symbol_factory.BitVecVal(2, 256)

    # Act
    solver.add(expression)
    result = solver.check()
    model = solver.model()

    x_concrete = model.eval(x.raw).as_long()

    # Assert
    assert z3.sat == result
    assert 2 == x_concrete



from typing import Dict, List
from typing_extensions import TypedDict


class AccountData(TypedDict):
    balance: str
    code: str
    nonce: int
    storage: dict


class InitialState(TypedDict):
    accounts: Dict[str, AccountData]


class TransactionData(TypedDict):
    address: str
    blockCoinbase: str
    blockDifficulty: str
    blockGasLimit: str
    blockNumber: str
    blockTime: str
    calldata: str
    gasLimit: str
    gasPrice: str
    input: str
    name: str
    origin: str
    value: str


class ConcreteData(TypedDict):
    initialState: InitialState
    steps: List[TransactionData]



"""The Mythril function signature database."""
import functools
import logging
import multiprocessing
import os
import sqlite3
import time
from collections import defaultdict
from typing import List, Set, DefaultDict, Dict

from mythril.ethereum.util import get_solc_json

log = logging.getLogger(__name__)

lock = multiprocessing.Lock()


def synchronized(sync_lock):
    """A decorator synchronizing multi-process access to a resource."""

    def wrapper(f):
        """The decorator's core function.

        :param f:
        :return:
        """

        @functools.wraps(f)
        def inner_wrapper(*args, **kw):
            """

            :param args:
            :param kw:
            :return:
            """
            with sync_lock:
                return f(*args, **kw)

        return inner_wrapper

    return wrapper


class Singleton(type):
    """A metaclass type implementing the singleton pattern."""

    _instances = dict()  # type: Dict[Singleton, Singleton]

    @synchronized(lock)
    def __call__(cls, *args, **kwargs):
        """Delegate the call to an existing resource or a a new one.

        This is not thread- or process-safe by default. It must be protected with
        a lock.

        :param args:
        :param kwargs:
        :return:
        """
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)

        return cls._instances[cls]


try:
    # load if available but do not fail
    import ethereum_input_decoder
    from ethereum_input_decoder.decoder import FourByteDirectoryOnlineLookupError
except ImportError:
    # fake it :)
    ethereum_input_decoder = None
    FourByteDirectoryOnlineLookupError = Exception


class SQLiteDB(object):
    """Simple context manager for sqlite3 databases.

    Commits everything at exit.
    """

    def __init__(self, path):
        """

        :param path:
        """
        self.path = path
        self.conn = None
        self.cursor = None

    def __enter__(self):
        """

        :return:
        """
        try:
            self.conn = sqlite3.connect(self.path)
        except sqlite3.OperationalError as e:
            raise sqlite3.OperationalError(f"Unable to Connect to path {self.path}")
        self.cursor = self.conn.cursor()
        return self.cursor

    def __exit__(self, exc_class, exc, traceback):
        """

        :param exc_class:
        :param exc:
        :param traceback:
        """
        self.conn.commit()
        self.conn.close()

    def __repr__(self):
        return "<SQLiteDB path={}>".format(self.path)


class SignatureDB(object, metaclass=Singleton):
    """"""

    def __init__(self, enable_online_lookup: bool = False, path: str = None) -> None:
        """
        :param enable_online_lookup:
        :param path:
        """
        self.enable_online_lookup = enable_online_lookup
        self.online_lookup_miss = set()  # type: Set[str]
        self.online_lookup_timeout = 0

        # if we're analysing a Solidity file, store its hashes
        # here to prevent unnecessary lookups
        self.solidity_sigs = defaultdict(list)  # type: DefaultDict[str, List[str]]
        if path is None:
            self.path = os.environ.get("MYTHRIL_DIR") or os.path.join(
                os.path.expanduser("~"), ".mythril"
            )
        self.path = os.path.join(self.path, "signatures.db")

        log.info("Using signature database at %s", self.path)
        # NOTE: Creates a new DB file if it doesn't exist already
        with SQLiteDB(self.path) as cur:
            cur.execute(
                (
                    "CREATE TABLE IF NOT EXISTS signatures"
                    "(byte_sig VARCHAR(10), text_sig VARCHAR(255),"
                    "PRIMARY KEY (byte_sig, text_sig))"
                )
            )

    def __getitem__(self, item: str) -> List[str]:
        """Provide dict interface db[sighash]

        :param item: 4-byte signature string
        :return: list of matching text signature strings
        """
        return self.get(byte_sig=item)

    @staticmethod
    def _normalize_byte_sig(byte_sig: str) -> str:
        """Adds a leading 0x to the byte signature if it's not already there.

        :param byte_sig: 4-byte signature string
        :return: normalized byte signature string
        """
        if not byte_sig.startswith("0x"):
            byte_sig = "0x" + byte_sig
        if not len(byte_sig) == 10:
            raise ValueError(
                "Invalid byte signature %s, must have 10 characters", byte_sig
            )
        return byte_sig

    def add(self, byte_sig: str, text_sig: str) -> None:
        """
        Adds a new byte - text signature pair to the database.
        :param byte_sig: 4-byte signature string
        :param text_sig: resolved text signature
        :return:
        """
        byte_sig = self._normalize_byte_sig(byte_sig)
        with SQLiteDB(self.path) as cur:
            # ignore new row if it's already in the DB (and would cause a unique constraint error)
            cur.execute(
                "INSERT OR IGNORE INTO signatures (byte_sig, text_sig) VALUES (?,?)",
                (byte_sig, text_sig),
            )

    def get(self, byte_sig: str, online_timeout: int = 2) -> List[str]:
        """Get a function text signature for a byte signature 1) try local
        cache 2) try online lookup (if enabled; if not flagged as unavailable)

        :param byte_sig: function signature hash as hexstr
        :param online_timeout: online lookup timeout
        :return: list of matching function text signatures
        """

        byte_sig = self._normalize_byte_sig(byte_sig)

        # check if we have any Solidity signatures to look up
        text_sigs = self.solidity_sigs.get(byte_sig)
        if text_sigs is not None:
            return text_sigs

        # try lookup in the local DB
        with SQLiteDB(self.path) as cur:
            cur.execute("SELECT text_sig FROM signatures WHERE byte_sig=?", (byte_sig,))
            text_sigs = cur.fetchall()

        if text_sigs:
            return [t[0] for t in text_sigs]

        # abort if we're not allowed to check 4byte or we already missed
        # the signature, or we're on a timeout
        if (
            not self.enable_online_lookup
            or byte_sig in self.online_lookup_miss
            or time.time() < self.online_lookup_timeout
        ):
            return []

        try:
            text_sigs = self.lookup_online(byte_sig=byte_sig, timeout=online_timeout)
            if not text_sigs:
                self.online_lookup_miss.add(byte_sig)
                return []
            else:
                for resolved in text_sigs:
                    self.add(byte_sig, resolved)
                return text_sigs
        except FourByteDirectoryOnlineLookupError as fbdole:
            # wait at least 2 mins to try again
            self.online_lookup_timeout = int(time.time()) + 2 * 60
            log.warning("Online lookup failed, not retrying for 2min: %s", fbdole)
            return []

    def import_solidity_file(
        self, file_path: str, solc_binary: str = "solc", solc_settings_json: str = None
    ):
        """Import Function Signatures from solidity source files.

        :param solc_binary:
        :param solc_settings_json:
        :param file_path: solidity source code file path
        :return:
        """
        solc_json = get_solc_json(file_path, solc_binary, solc_settings_json)
        self.add_sigs(file_path, solc_json)

    def add_sigs(self, file_path: str, solc_json):
        for contract in solc_json["contracts"][file_path].values():
            if "methodIdentifiers" not in contract["evm"]:
                continue
            for name, hash_ in contract["evm"]["methodIdentifiers"].items():
                sig = "0x{}".format(hash_)
                if sig in self.solidity_sigs:
                    self.solidity_sigs[sig].append(name)
                else:
                    self.solidity_sigs[sig] = [name]
                self.add(sig, name)

    @staticmethod
    def lookup_online(byte_sig: str, timeout: int, proxies=None) -> List[str]:
        """Lookup function signatures from 4byte.directory.

        :param byte_sig: function signature hash as hexstr
        :param timeout: optional timeout for online lookup
        :param proxies: optional proxy servers for online lookup
        :return: a list of matching function signatures for this hash
        """
        if not ethereum_input_decoder:
            return []
        return list(
            ethereum_input_decoder.decoder.FourByteDirectory.lookup_signatures(
                byte_sig, timeout=timeout, proxies=proxies
            )
        )

    def __repr__(self):
        return "<SignatureDB path='{}' enable_online_lookup={}>".format(
            self.path, self.enable_online_lookup
        )



import json
import binascii

from copy import deepcopy
from datetime import datetime
from typing import Dict, List, Tuple

from mythril.concolic.concrete_data import ConcreteData

from mythril.disassembler.disassembly import Disassembly
from mythril.laser.ethereum.svm import LaserEVM
from mythril.laser.ethereum.state.world_state import WorldState
from mythril.laser.ethereum.state.account import Account
from mythril.laser.ethereum.time_handler import time_handler
from mythril.laser.ethereum.transaction.concolic import execute_transaction
from mythril.laser.plugin.loader import LaserPluginLoader
from mythril.laser.smt import Expression, BitVec, symbol_factory
from mythril.laser.ethereum.transaction.transaction_models import tx_id_manager
from mythril.plugin.discovery import PluginDiscovery


def setup_concrete_initial_state(concrete_data: ConcreteData) -> WorldState:
    """
    Sets up concrete initial state
    :param concrete_data: Concrete data
    :return: initialised world state
    """
    world_state = WorldState()
    for address, details in concrete_data["initialState"]["accounts"].items():
        account = Account(address, concrete_storage=True)
        account.code = Disassembly(details["code"][2:])
        account.nonce = details["nonce"]
        if type(details["storage"]) == str:
            details["storage"] = eval(details["storage"])  # type: ignore
        for key, value in details["storage"].items():
            key_bitvec = symbol_factory.BitVecVal(int(key, 16), 256)
            account.storage[key_bitvec] = symbol_factory.BitVecVal(int(value, 16), 256)

        world_state.put_account(account)
        account.set_balance(int(details["balance"], 16))
    return world_state


def concrete_execution(concrete_data: ConcreteData) -> Tuple[WorldState, List]:
    """
    Executes code concretely to find the path to be followed by concolic executor
    :param concrete_data: Concrete data
    :return: path trace
    """
    tx_id_manager.restart_counter()
    init_state = setup_concrete_initial_state(concrete_data)
    laser_evm = LaserEVM(execution_timeout=1000)
    laser_evm.open_states = [deepcopy(init_state)]
    plugin_loader = LaserPluginLoader()
    assert PluginDiscovery().is_installed("myth_concolic_execution")
    trace_plugin = PluginDiscovery().installed_plugins["myth_concolic_execution"]()
    time_handler.start_execution(laser_evm.execution_timeout)

    plugin_loader.load(trace_plugin)
    laser_evm.time = datetime.now()
    plugin_loader.instrument_virtual_machine(laser_evm, None)
    for transaction in concrete_data["steps"]:
        execute_transaction(
            laser_evm,
            callee_address=transaction["address"],
            caller_address=symbol_factory.BitVecVal(
                int(transaction["origin"], 16), 256
            ),
            origin_address=symbol_factory.BitVecVal(
                int(transaction["origin"], 16), 256
            ),
            gas_limit=int(transaction.get("gasLimit", "0x9999999999999999999999"), 16),
            data=binascii.a2b_hex(transaction["input"][2:]),
            gas_price=int(transaction.get("gasPrice", "0x773594000"), 16),
            value=int(transaction["value"], 16),
            track_gas=False,
        )

    tx_id_manager.restart_counter()
    return init_state, plugin_loader.plugin_list["MythX Trace Finder"].tx_trace  # type: ignore



import os
import time
import errno

"""
credits: https://github.com/dmfrey/FileLock
"""


class LockFileException(Exception):
    pass


class LockFile(object):
    """
    Locks files.
    """

    def __init__(self, file_name, timeout=100, delay=0.05):
        """
        Initialises the file locker
        """
        if timeout is not None and delay is None:
            raise ValueError("If timeout is not None, then delay must not be None.")
        self.is_locked = False
        self.lockfile = os.path.join(os.getcwd(), f"{file_name}.lock")
        self.file_name = file_name
        self.timeout = timeout
        self.delay = delay

    def acquire(self):
        """
        Acquires a lock when possible.
        """
        start_time = time.time()
        while True:
            try:
                self.fd = os.open(self.lockfile, os.O_CREAT | os.O_EXCL | os.O_RDWR)
                self.is_locked = True
                break
            except OSError as e:
                if e.errno != errno.EEXIST:
                    raise
                if (time.time() - start_time) >= self.timeout:
                    raise FileLockException(
                        f"Stuck for more than {self.timeout} seconds waiting to unlock the file {self.filename}."
                    )
                time.sleep(self.delay)

    def release(self):
        """
        Releases the lock
        """
        if self.is_locked:
            os.close(self.fd)
            os.unlink(self.lockfile)
            self.is_locked = False

    def __enter__(self):
        """
        Lock gets acquired at the `with` statement.
        """
        if not self.is_locked:
            self.acquire()
        return self

    def __exit__(self, type, value, traceback):
        """
        Lock get's released at the end of the `with` block
        """
        if self.is_locked:
            self.release()

    def __del__(self):
        """
        Releases the lock during deletion.
        """
        self.release()



import pytest

from mythril.laser.ethereum.evm_exceptions import VmException, OutOfGasException
from mythril.disassembler.disassembly import Disassembly
from mythril.laser.ethereum.state.environment import Environment
from mythril.laser.ethereum.state.world_state import WorldState
from mythril.laser.ethereum.state.account import Account
from mythril.laser.ethereum.state.machine_state import MachineState
from mythril.laser.ethereum.state.global_state import GlobalState
from mythril.laser.ethereum.instructions import Instruction
from mythril.laser.ethereum.transaction.transaction_models import MessageCallTransaction
from mythril.laser.smt import symbol_factory, simplify


def get_state():
    world_state = WorldState()
    account = world_state.create_account(balance=10, address=101)
    account.code = Disassembly("0x60045e005c5d")
    environment = Environment(account, None, None, None, None, None, None)
    state = GlobalState(world_state, environment, None, MachineState(gas_limit=8000000))
    state.transaction_stack.append(
        (MessageCallTransaction(world_state=WorldState(), gas_limit=8000000), None)
    )
    return state


BVV = symbol_factory.BitVecVal
BV = symbol_factory.BitVecSym

test_data = (([BVV(5, 256)], [], 2, -1, ()),)


def test_jumpsub_success():
    # Arrange
    state = get_state()
    state.mstate.pc = 2
    state.mstate.stack = [BVV(4, 256)]
    state.mstate.subroutine_stack = []
    instruction = Instruction("jumpsub", dynamic_loader=None)

    # Act
    new_state = instruction.evaluate(state)[0]

    # Assert
    assert new_state.mstate.pc == 5
    assert new_state.mstate.stack == []
    assert new_state.mstate.subroutine_stack == [3]


def test_jumpsub_fail():
    # Arrange
    state = get_state()
    state.mstate.pc = 2
    state.mstate.stack = [BVV(5, 256)]
    state.mstate.subroutine_stack = []
    instruction = Instruction("jumpsub", dynamic_loader=None)

    # Act + Assert
    with pytest.raises(VmException):
        instruction.evaluate(state)[0]


def test_beginsub():
    # Arrange
    state = get_state()
    state.mstate.pc = 3
    state.mstate.stack = []
    state.mstate.subroutine_stack = []
    instruction = Instruction("beginsub", dynamic_loader=None)

    # Act + Assert
    with pytest.raises(OutOfGasException):
        instruction.evaluate(state)[0]


def test_returnsub():
    # Arrange
    state = get_state()
    state.mstate.pc = 5
    state.mstate.stack = []
    state.mstate.subroutine_stack = [3]
    instruction = Instruction("returnsub", dynamic_loader=None)

    # Act
    new_state = instruction.evaluate(state)[0]

    # Assert
    assert new_state.mstate.pc == 3



import logging
import re
from typing import Union, List, cast, Optional
from eth.constants import GAS_CALLSTIPEND

import mythril.laser.ethereum.util as util
from mythril.laser.ethereum.util import insert_ret_val
from mythril.laser.ethereum import natives
from mythril.laser.ethereum.instruction_data import calculate_native_gas
from mythril.laser.ethereum.state.account import Account
from mythril.laser.ethereum.natives import PRECOMPILE_COUNT, PRECOMPILE_FUNCTIONS
from mythril.laser.ethereum.state.calldata import (
    BaseCalldata,
    SymbolicCalldata,
    ConcreteCalldata,
)
from mythril.laser.ethereum.state.global_state import GlobalState
from mythril.laser.smt import BitVec, If
from mythril.laser.smt import simplify, Expression, symbol_factory
from mythril.support.loader import DynLoader


class hevm_cheat_code:
    # https://github.com/dapphub/ds-test/blob/cd98eff28324bfac652e63a239a60632a761790b/src/test.sol

    address = 0x7109709ECFA91A80626FF3989D68F67F5B1DD12D

    fail_payload = int(
        "70ca10bb"
        + "0000000000000000000000007109709ecfa91a80626ff3989d68f67f5b1dd12d"
        + "6661696c65640000000000000000000000000000000000000000000000000000"
        + "0000000000000000000000000000000000000000000000000000000000000001",
        16,
    )

    assume_sig = 0x4C63E562

    @staticmethod
    def is_cheat_address(address):
        if int(address, 16) == int("0x7109709ECfa91a80626fF3989D68f67F5b1DD12D", 16):
            return True
        if int(address, 16) == int("0x72c68108a82e82617b93d1be0d7975d762035015", 16):
            return True
        return False


def handle_cheat_codes(
    global_state: GlobalState,
    callee_address: Union[str, BitVec],
    call_data: BaseCalldata,
    memory_out_offset: Union[int, Expression],
    memory_out_size: Union[int, Expression],
):

    insert_ret_val(global_state)
    pass



"""This module contains account-related functionality.

This includes classes representing accounts and their storage.
"""
import logging
from copy import copy, deepcopy
from typing import Any, Dict, Union, Set


from mythril.laser.smt import Array, K, BitVec, simplify, BaseArray, If, Bool
from mythril.disassembler.disassembly import Disassembly
from mythril.laser.smt import symbol_factory
from mythril.support.support_args import args

log = logging.getLogger(__name__)


class Storage:
    """Storage class represents the storage of an Account."""

    def __init__(self, concrete=False, address=None, dynamic_loader=None) -> None:
        """Constructor for Storage.

        :param concrete: bool indicating whether to interpret uninitialized storage as concrete versus symbolic
        """
        if concrete and args.unconstrained_storage is False:
            self._standard_storage: BaseArray = K(256, 256, 0)
        else:
            self._standard_storage = Array(f"Storage{address}", 256, 256)

        self.printable_storage: Dict[BitVec, BitVec] = {}

        self.dynld = dynamic_loader
        self.storage_keys_loaded: Set[int] = set()
        self.address = address

        # Stores all keys set in the storage
        self.keys_set: Set[BitVec] = set()

        # Stores all get keys in the storage
        self.keys_get: Set[BitVec] = set()

    def __getitem__(self, item: BitVec) -> BitVec:
        storage = self._standard_storage
        self.keys_get.add(item)
        if (
            self.address
            and self.address.value != 0
            and item.symbolic is False
            and int(item.value) not in self.storage_keys_loaded
            and (self.dynld and self.dynld.active)
            and args.unconstrained_storage is False
        ):
            try:
                value = symbol_factory.BitVecVal(
                    int(
                        self.dynld.read_storage(
                            contract_address="0x{:040X}".format(self.address.value),
                            index=int(item.value),
                        ),
                        16,
                    ),
                    256,
                )

                for key in self.keys_set:
                    value = If(key == item, storage[item], value)

                storage[item] = value
                self.storage_keys_loaded.add(int(item.value))
                self.printable_storage[item] = storage[item]
            except ValueError as e:
                log.debug("Couldn't read storage at %s: %s", item, e)

        return simplify(storage[item])

    def __setitem__(self, key, value: Any) -> None:
        if isinstance(value, Bool):
            value = If(value, 1, 0)

        self.printable_storage[key] = value
        self._standard_storage[key] = value

        self.keys_set.add(key)

        if key.symbolic is False:
            self.storage_keys_loaded.add(int(key.value))

    def __deepcopy__(self, memodict=dict()):
        concrete = isinstance(self._standard_storage, K)
        storage = Storage(
            concrete=concrete, address=self.address, dynamic_loader=self.dynld
        )
        storage._standard_storage = deepcopy(self._standard_storage)
        storage.printable_storage = copy(self.printable_storage)
        storage.storage_keys_loaded = copy(self.storage_keys_loaded)
        storage.keys_set = deepcopy(self.keys_set)
        storage.keys_get = deepcopy(self.keys_get)
        return storage

    def __str__(self) -> str:
        # TODO: Do something better here
        return str(self.printable_storage)


class Account:
    """Account class representing ethereum accounts."""

    def __init__(
        self,
        address: Union[BitVec, str],
        code=None,
        contract_name=None,
        balances: Array = None,
        concrete_storage=False,
        dynamic_loader=None,
        nonce=0,
    ) -> None:
        """Constructor for account.

        :param address: Address of the account
        :param code: The contract code of the account
        :param contract_name: The name associated with the account
        :param balances: The balance for the account
        :param concrete_storage: Interpret storage as concrete
        """
        self.concrete_storage = concrete_storage
        self.nonce = nonce
        self.code = code or Disassembly("")
        self.address = (
            address
            if isinstance(address, BitVec)
            else symbol_factory.BitVecVal(int(address, 16), 256)
        )

        self.storage = Storage(
            concrete_storage, address=self.address, dynamic_loader=dynamic_loader
        )

        # Metadata
        if contract_name is None:
            self.contract_name = (
                "{0:#0{1}x}".format(self.address.value, 42)
                if not self.address.symbolic
                else "unknown"
            )
        else:
            self.contract_name = contract_name

        self.deleted = False

        self._balances = balances
        self.balance = lambda: self._balances[self.address]

    def __str__(self) -> str:
        return str(self.as_dict)

    def set_balance(self, balance: Union[int, BitVec]) -> None:
        """

        :param balance:
        """
        balance = (
            symbol_factory.BitVecVal(balance, 256)
            if isinstance(balance, int)
            else balance
        )
        assert self._balances is not None
        self._balances[self.address] = balance

    def set_storage(self, storage: Dict):
        """
        Sets concrete storage
        """
        for key, value in storage.items():
            concrete_key, concrete_value = int(key, 16), int(value, 16)
            self.storage[
                symbol_factory.BitVecVal(concrete_key, 256)
            ] = symbol_factory.BitVecVal(concrete_value, 256)

    def add_balance(self, balance: Union[int, BitVec]) -> None:
        """

        :param balance:
        """
        balance = (
            symbol_factory.BitVecVal(balance, 256)
            if isinstance(balance, int)
            else balance
        )
        self._balances[self.address] += balance

    @property
    def as_dict(self) -> Dict:
        """

        :return:
        """
        return {
            "nonce": self.nonce,
            "code": self.serialised_code(),
            "balance": self.balance(),
            "storage": self.storage,
        }

    def serialised_code(self):
        if type(self.code.bytecode) == str:
            return self.code.bytecode
        new_code = "0x"
        for byte in self.code.bytecode:
            if type(byte) == int:
                new_code += hex(byte)
            else:
                new_code += "<call_data>"
        return new_code

    def __copy__(self, memodict={}):
        new_account = Account(
            address=self.address,
            code=self.code,
            contract_name=self.contract_name,
            balances=deepcopy(self._balances),
            concrete_storage=self.concrete_storage,
            nonce=self.nonce,
        )
        new_account.storage = deepcopy(self.storage)
        new_account.code = self.code
        return new_account



from abc import ABC, abstractmethod
from typing import List
from mythril.laser.ethereum.state.global_state import GlobalState


class BasicSearchStrategy(ABC):
    """
    A basic search strategy which halts based on depth
    """

    def __init__(self, work_list, max_depth, **kwargs):
        self.work_list = work_list  # type: List[GlobalState]
        self.max_depth = max_depth

    def __iter__(self):
        return self

    @abstractmethod
    def get_strategic_global_state(self):
        """"""
        raise NotImplementedError("Must be implemented by a subclass")

    def run_check(self):
        return True

    def __next__(self):
        try:
            global_state = self.get_strategic_global_state()
            if global_state.mstate.depth >= self.max_depth:
                return self.__next__()
            return global_state
        except (IndexError, StopIteration):
            raise StopIteration


class CriterionSearchStrategy(BasicSearchStrategy):
    """
    If a criterion is satisfied, the search halts
    """

    def __init__(self, work_list, max_depth, **kwargs):
        super().__init__(work_list, max_depth, **kwargs)
        self._satisfied_criterion = False

    def get_strategic_global_state(self):
        if self._satisfied_criterion:
            raise StopIteration
        try:
            return self.get_strategic_global_state()
        except StopIteration:
            raise StopIteration

    def set_criterion_satisfied(self):
        self._satisfied_criterion = True



"""This module declares classes to represent call data."""
from typing import cast, Union, Tuple, List


from typing import Any, Union

from z3 import Model, unsat, unknown
from z3.z3types import Z3Exception

from mythril.laser.ethereum.util import get_concrete_int

from mythril.laser.smt import (
    Array,
    BitVec,
    Bool,
    Concat,
    Expression,
    If,
    K,
    simplify,
    symbol_factory,
    Solver,
)


class BaseCalldata:
    """Base calldata class This represents the calldata provided when sending a
    transaction to a contract."""

    def __init__(self, tx_id: str) -> None:
        """

        :param tx_id:
        """
        self.tx_id = tx_id

    @property
    def calldatasize(self) -> BitVec:
        """

        :return: Calldata size for this calldata object
        """
        result = self.size
        if isinstance(result, int):
            return symbol_factory.BitVecVal(result, 256)
        return result

    def get_word_at(self, offset: int) -> Expression:
        """Gets word at offset.

        :param offset:
        :return:
        """
        parts = self[offset : offset + 32]
        return simplify(Concat(parts))

    def __getitem__(self, item: Union[int, slice, BitVec]) -> Any:
        """

        :param item:
        :return:
        """
        if isinstance(item, int) or isinstance(item, Expression):
            return self._load(item)

        if isinstance(item, slice):
            start = 0 if item.start is None else item.start
            step = 1 if item.step is None else item.step
            stop = self.size if item.stop is None else item.stop

            try:
                current_index = (
                    start
                    if isinstance(start, BitVec)
                    else symbol_factory.BitVecVal(start, 256)
                )
                parts = []
                while True:
                    s = Solver()
                    s.set_timeout(1000)
                    s.add(current_index != stop)
                    result = s.check()
                    if result in (unsat, unknown):
                        break
                    element = self._load(current_index)
                    if not isinstance(element, Expression):
                        element = symbol_factory.BitVecVal(element, 8)

                    parts.append(element)
                    current_index = simplify(current_index + step)

            except Z3Exception:
                raise IndexError("Invalid Calldata Slice")
            return parts

        raise ValueError

    def _load(self, item: Union[int, BitVec]) -> Any:
        """

        :param item:
        """
        raise NotImplementedError()

    @property
    def size(self) -> Union[BitVec, int]:
        """Returns the exact size of this calldata, this is not normalized.

        :return: unnormalized call data size
        """
        raise NotImplementedError()

    def concrete(self, model: Model) -> list:
        """Returns a concrete version of the calldata using the provided model.

        :param model:
        """
        raise NotImplementedError


class ConcreteCalldata(BaseCalldata):
    """A concrete call data representation."""

    def __init__(self, tx_id: str, calldata: list) -> None:
        """Initializes the ConcreteCalldata object.

        :param tx_id: Id of the transaction that the calldata is for.
        :param calldata: The concrete calldata content
        """
        self._concrete_calldata = calldata
        self._calldata = K(256, 8, 0)
        for i, element in enumerate(calldata, 0):
            element = (
                symbol_factory.BitVecVal(element, 8)
                if isinstance(element, int)
                else element
            )
            self._calldata[symbol_factory.BitVecVal(i, 256)] = element

        super().__init__(tx_id)

    def _load(self, item: Union[int, BitVec]) -> BitVec:
        """

        :param item:
        :return:
        """
        item = symbol_factory.BitVecVal(item, 256) if isinstance(item, int) else item
        return simplify(self._calldata[item])

    def concrete(self, model: Model) -> list:
        """

        :param model:
        :return:
        """
        return self._concrete_calldata

    @property
    def size(self) -> int:
        """

        :return:
        """
        return len(self._concrete_calldata)


class BasicConcreteCalldata(BaseCalldata):
    """A base class to represent concrete call data."""

    def __init__(self, tx_id: str, calldata: list) -> None:
        """Initializes the ConcreteCalldata object, that doesn't use z3 arrays.

        :param tx_id: Id of the transaction that the calldata is for.
        :param calldata: The concrete calldata content
        """
        self._calldata = calldata
        super().__init__(tx_id)

    def _load(self, item: Union[int, Expression]) -> Any:
        """

        :param item:
        :return:
        """
        if isinstance(item, int):
            try:
                return self._calldata[item]
            except IndexError:
                return 0

        value = symbol_factory.BitVecVal(0x0, 8)
        for i in range(self.size):
            value = If(cast(Union[BitVec, Bool], item) == i, self._calldata[i], value)
        return value

    def concrete(self, model: Model) -> list:
        """

        :param model:
        :return:
        """
        concrete_calldata = []
        for data in self._calldata:
            if isinstance(data, BitVec) and data.symbolic and model is not None:
                concrete_calldata.append(model.eval(data, model_completion=True))
            elif isinstance(data, BitVec) and data.symbolic is False:
                concrete_calldata.append(data)
            else:
                break
        return concrete_calldata

    @property
    def size(self) -> int:
        """

        :return:
        """
        return len(self._calldata)


class SymbolicCalldata(BaseCalldata):
    """A class for representing symbolic call data."""

    def __init__(self, tx_id: str) -> None:
        """Initializes the SymbolicCalldata object.

        :param tx_id: Id of the transaction that the calldata is for.
        """
        self._size = symbol_factory.BitVecSym(str(tx_id) + "_calldatasize", 256)
        self._calldata = Array("{}_calldata".format(tx_id), 256, 8)
        super().__init__(tx_id)

    def _load(self, item: Union[int, BitVec]) -> Any:
        """

        :param item:
        :return:
        """
        item = symbol_factory.BitVecVal(item, 256) if isinstance(item, int) else item
        return simplify(
            If(
                item < self._size,
                simplify(self._calldata[cast(BitVec, item)]),
                symbol_factory.BitVecVal(0, 8),
            )
        )

    def concrete(self, model: Model) -> list:
        """

        :param model:
        :return:
        """
        concrete_length = model.eval(self.size.raw, model_completion=True).as_long()
        result = []
        for i in range(concrete_length):
            value = self._load(i)
            c_value = model.eval(value.raw, model_completion=True).as_long()
            result.append(c_value)

        return result

    @property
    def size(self) -> BitVec:
        """

        :return:
        """
        return self._size


class BasicSymbolicCalldata(BaseCalldata):
    """A basic class representing symbolic call data."""

    def __init__(self, tx_id: str) -> None:
        """Initializes the SymbolicCalldata object.

        :param tx_id: Id of the transaction that the calldata is for.
        """
        self._reads = []  # type: List[Tuple[Union[int, BitVec], BitVec]]
        self._size = symbol_factory.BitVecSym(str(tx_id) + "_calldatasize", 256)
        super().__init__(tx_id)

    def _load(self, item: Union[int, BitVec], clean=False) -> Any:
        expr_item = (
            symbol_factory.BitVecVal(item, 256) if isinstance(item, int) else item
        )  # type: BitVec

        symbolic_base_value = If(
            expr_item >= self._size,
            symbol_factory.BitVecVal(0, 8),
            BitVec(
                symbol_factory.BitVecSym(
                    "{}_calldata_{}".format(self.tx_id, str(item)), 8
                )
            ),
        )
        return_value = symbolic_base_value
        for r_index, r_value in self._reads:
            return_value = If(r_index == expr_item, r_value, return_value)
        if not clean:
            self._reads.append((expr_item, symbolic_base_value))
        return simplify(return_value)

    def concrete(self, model: Model) -> list:
        """

        :param model:
        :return:
        """
        concrete_length = get_concrete_int(model.eval(self.size, model_completion=True))
        result = []
        for i in range(concrete_length):
            value = self._load(i, clean=True)
            c_value = get_concrete_int(model.eval(value, model_completion=True))
            result.append(c_value)

        return result

    @property
    def size(self) -> BitVec:
        """

        :return:
        """
        return self._size



from time import time

from mythril.support.support_utils import Singleton

from typing import Callable


def stat_smt_query(func: Callable):
    """Measures statistics for annotated smt query check function"""
    stat_store = SolverStatistics()

    def function_wrapper(*args, **kwargs):
        if not stat_store.enabled:
            return func(*args, **kwargs)

        stat_store.query_count += 1
        begin = time()

        result = func(*args, **kwargs)

        end = time()
        stat_store.solver_time += end - begin

        return result

    return function_wrapper


class SolverStatistics(object, metaclass=Singleton):
    """Solver Statistics Class

    Keeps track of the important statistics around smt queries
    """

    def __init__(self):
        self.enabled = False
        self.query_count = 0
        self.solver_time = 0

    def __repr__(self):
        return "Query count: {} \nSolver time: {}".format(
            self.query_count, self.solver_time
        )



"""Don't ask."""
#!/usr/bin/env python
#
# "THE BEER-WARE LICENSE" (Revision 43~maze)
#
# <maze@pyth0n.org> wrote these files. As long as you retain this notice you
# can do whatever you want with this stuff. If we meet some day, and you think
# this stuff is worth it, you can buy me a beer in return.

# https://github.com/tehmaze/lolcat

import atexit
import math
import os
import random
import re
import sys
import time
import argparse

PY3 = sys.version_info >= (3,)

# Reset terminal colors at exit
def reset():
    """"""
    sys.stdout.write("\x1b[0m")
    sys.stdout.flush()


atexit.register(reset)


STRIP_ANSI = re.compile(r"\x1b\[(\d+)(;\d+)?(;\d+)?[m|K]")
COLOR_ANSI = (
    (0x00, 0x00, 0x00),
    (0xCD, 0x00, 0x00),
    (0x00, 0xCD, 0x00),
    (0xCD, 0xCD, 0x00),
    (0x00, 0x00, 0xEE),
    (0xCD, 0x00, 0xCD),
    (0x00, 0xCD, 0xCD),
    (0xE5, 0xE5, 0xE5),
    (0x7F, 0x7F, 0x7F),
    (0xFF, 0x00, 0x00),
    (0x00, 0xFF, 0x00),
    (0xFF, 0xFF, 0x00),
    (0x5C, 0x5C, 0xFF),
    (0xFF, 0x00, 0xFF),
    (0x00, 0xFF, 0xFF),
    (0xFF, 0xFF, 0xFF),
)


class LolCat(object):
    """Cats lel."""

    def __init__(self, mode=256, output=sys.stdout):
        self.mode = mode
        self.output = output

    def _distance(self, rgb1, rgb2):
        return sum(map(lambda c: (c[0] - c[1]) ** 2, zip(rgb1, rgb2)))

    def ansi(self, rgb):
        """

        :param rgb:
        :return:
        """
        r, g, b = rgb

        if self.mode in (8, 16):
            colors = COLOR_ANSI[: self.mode]
            matches = [
                (self._distance(c, map(int, rgb)), i) for i, c in enumerate(colors)
            ]
            matches.sort()
            color = matches[0][1]

            return "3%d" % (color,)
        else:
            gray_possible = True
            sep = 2.5

            while gray_possible:
                if r < sep or g < sep or b < sep:
                    gray = r < sep and g < sep and b < sep
                    gray_possible = False

                sep += 42.5

            if gray:
                color = 232 + int(float(sum(rgb) / 33.0))
            else:
                color = sum(
                    [16]
                    + [
                        int(6 * float(val) / 256) * mod
                        for val, mod in zip(rgb, [36, 6, 1])
                    ]
                )

            return "38;5;%d" % (color,)

    def wrap(self, *codes):
        """

        :param codes:
        :return:
        """
        return "\x1b[%sm" % ("".join(codes),)

    def rainbow(self, freq, i):
        """

        :param freq:
        :param i:
        :return:
        """
        r = math.sin(freq * i) * 127 + 128
        g = math.sin(freq * i + 2 * math.pi / 3) * 127 + 128
        b = math.sin(freq * i + 4 * math.pi / 3) * 127 + 128
        return [r, g, b]

    def cat(self, fd, options):
        """

        :param fd:
        :param options:
        """
        if options.animate:
            self.output.write("\x1b[?25l")

        for line in fd:
            options.os += 1
            self.println(line, options)

        if options.animate:
            self.output.write("\x1b[?25h")

    def println(self, s, options):
        """

        :param s:
        :param options:
        """
        s = s.rstrip()
        if options.force or self.output.isatty():
            s = STRIP_ANSI.sub("", s)

        if options.animate:
            self.println_ani(s, options)
        else:
            self.println_plain(s, options)

        self.output.write("\n")
        self.output.flush()

    def println_ani(self, s, options):
        """

        :param s:
        :param options:
        :return:
        """
        if not s:
            return

        for _ in range(1, options.duration):
            self.output.write("\x1b[%dD" % (len(s),))
            self.output.flush()
            options.os += options.spread
            self.println_plain(s, options)
            time.sleep(1.0 / options.speed)

    def println_plain(self, s, options):
        """

        :param s:
        :param options:
        """
        for i, c in enumerate(s if PY3 else s.decode(options.charset_py2, "replace")):
            rgb = self.rainbow(options.freq, options.os + i / options.spread)
            self.output.write(
                "".join(
                    [
                        self.wrap(self.ansi(rgb)),
                        c if PY3 else c.encode(options.charset_py2, "replace"),
                    ]
                )
            )


def detect_mode(term_hint="xterm-256color"):
    """Poor-mans color mode detection."""
    if "ANSICON" in os.environ:
        return 16
    elif os.environ.get("ConEmuANSI", "OFF") == "ON":
        return 256
    else:
        term = os.environ.get("TERM", term_hint)
        if term.endswith("-256color") or term in ("xterm", "screen"):
            return 256
        elif term.endswith("-color") or term in ("rxvt",):
            return 16
        else:
            return 256  # optimistic default


def run():
    """Main entry point."""

    parser = argparse.ArgumentParser(usage=r"%prog [<options>] [file ...]")
    parser.add_argument(
        "-p", "--spread", type=float, default=3.0, help="Rainbow spread"
    )
    parser.add_argument(
        "-F", "--freq", type=float, default=0.1, help="Rainbow frequency"
    )
    parser.add_argument("-S", "--seed", type=int, default=0, help="Rainbow seed")
    parser.add_argument(
        "-a",
        "--animate",
        action="store_true",
        default=False,
        help="Enable psychedelics",
    )
    parser.add_argument(
        "-d", "--duration", type=int, default=12, help="Animation duration"
    )
    parser.add_argument(
        "-s", "--speed", type=float, default=20.0, help="Animation speed"
    )
    parser.add_argument(
        "-f",
        "--force",
        action="store_true",
        default=False,
        help="Force colour even when stdout is not a tty",
    )

    parser.add_argument(
        "-3", action="store_const", dest="mode", const=8, help="Force 3 bit colour mode"
    )
    parser.add_argument(
        "-4",
        action="store_const",
        dest="mode",
        const=16,
        help="Force 4 bit colour mode",
    )
    parser.add_argument(
        "-8",
        action="store_const",
        dest="mode",
        const=256,
        help="Force 8 bit colour mode",
    )

    parser.add_argument(
        "-c",
        "--charset-py2",
        default="utf-8",
        help="Manually set a charset to convert from, for python 2.7",
    )

    options = parser.parse_args()
    options.os = random.randint(0, 256) if options.seed == 0 else options.seed
    options.mode = options.mode or detect_mode()

    lolcat = LolCat(mode=options.mode)
    args = ["-"]

    for filename in args:
        if filename == "-":
            lolcat.cat(sys.stdin, options)
        else:
            try:
                with open(filename, "r") as handle:
                    lolcat.cat(handle, options)
            except IOError as error:
                sys.stderr.write(str(error) + "\n")


if __name__ == "__main__":
    sys.exit(run())



"""This module contains a basic Ethereum RPC client.

This code is adapted from: https://github.com/ConsenSys/ethjsonrpc
"""
import json
import logging

import requests
from requests.adapters import HTTPAdapter
from requests.exceptions import ConnectionError as RequestsConnectionError

from .base_client import BaseClient
from .exceptions import (
    BadJsonError,
    BadResponseError,
    BadStatusCodeError,
    ConnectionError,
)

log = logging.getLogger(__name__)

GETH_DEFAULT_RPC_PORT = 8545
ETH_DEFAULT_RPC_PORT = 8545
PARITY_DEFAULT_RPC_PORT = 8545
PYETHAPP_DEFAULT_RPC_PORT = 4000
MAX_RETRIES = 3
JSON_MEDIA_TYPE = "application/json"


class EthJsonRpc(BaseClient):
    """Ethereum JSON-RPC client class."""

    def __init__(self, host="localhost", port=GETH_DEFAULT_RPC_PORT, tls=False):
        """

        :param host:
        :param port:
        :param tls:
        """
        self.host = host
        self.port = port
        self.tls = tls
        self.session = requests.Session()
        self.session.mount(self.host, HTTPAdapter(max_retries=MAX_RETRIES))

    def _call(self, method, params=None, _id=1):
        """

        :param method:
        :param params:
        :param _id:
        :return:
        """
        params = params or []
        data = {"jsonrpc": "2.0", "method": method, "params": params, "id": _id}
        scheme = "http"
        if self.tls:
            scheme += "s"
        if self.host:
            if self.port:
                url = "{}://{}:{}".format(scheme, self.host, self.port)
            else:
                url = "{}://{}".format(scheme, self.host)

        else:
            url = "{}".format(scheme)

        headers = {"Content-Type": JSON_MEDIA_TYPE}
        log.debug("rpc send: %s" % json.dumps(data))
        try:
            r = self.session.post(url, headers=headers, data=json.dumps(data))
        except RequestsConnectionError:
            raise ConnectionError
        if r.status_code / 100 != 2:
            raise BadStatusCodeError(r.status_code)
        try:
            response = r.json()
            log.debug("rpc response: %s" % response)
        except ValueError:
            raise BadJsonError(r.text)
        try:
            return response["result"]
        except KeyError:
            raise BadResponseError(response)

    def close(self):
        """Close the RPC client's session."""
        self.session.close()



import pytest
from mythril.laser.smt import symbol_factory
from mythril.laser.ethereum.state.account import Storage
from mythril.laser.smt import Expression
from mythril.support.support_args import args

BVV = symbol_factory.BitVecVal

storage_uninitialized_test_data = [({}, 1), ({1: 5}, 2), ({1: 5, 3: 10}, 2)]


@pytest.mark.parametrize("initial_storage,key", storage_uninitialized_test_data)
def test_concrete_storage_uninitialized_index(initial_storage, key):
    # Arrange
    args.unconstrained_storage = False
    storage = Storage(concrete=True)
    for k, val in initial_storage.items():
        storage[BVV(k, 256)] = BVV(val, 256)

    # Act
    value = storage[BVV(key, 256)]

    # Assert
    assert value == 0


@pytest.mark.parametrize("initial_storage,key", storage_uninitialized_test_data)
def test_symbolic_storage_uninitialized_index(initial_storage, key):
    # Arrange
    storage = Storage(concrete=False)
    for k, val in initial_storage.items():
        storage[BVV(k, 256)] = BVV(val, 256)

    # Act
    value = storage[BVV(key, 256)]

    # Assert
    assert isinstance(value, Expression)


def test_storage_set_item():
    # Arrange
    storage = Storage()

    # Act
    storage[BVV(1, 256)] = BVV(13, 256)

    # Assert
    assert storage[BVV(1, 256)] == BVV(13, 256)


def test_storage_change_item():
    # Arrange
    storage = Storage()
    storage[BVV(1, 256)] = BVV(12, 256)
    # Act
    storage[BVV(1, 256)] = BVV(14, 256)

    # Assert
    assert storage[BVV(1, 256)] == BVV(14, 256)



# -*- coding: utf-8 -*-
"""install mythril and deploy source-dist and wheel to pypi.python.org.

deps (requires up2date version):
    *) pip install --upgrade pip wheel setuptools twine
publish to pypi w/o having to convert Readme.md to RST:
    1) #> python setup.py sdist bdist_wheel
    2) #> twine upload dist/*   #<specify bdist_wheel version to upload>; #optional --repository <testpypi> or  --repository-url <testpypi-url>
"""
from setuptools import setup, find_packages
from setuptools.command.install import install as _install
from subprocess import check_call

import sys
import os
import io


class InstallCommand(_install):
    def run(self):
        check_call([sys.executable, "-m", "pip", "install", "cython"])
        _install.run(self)


# Package meta-data.
NAME = "mythril"
DESCRIPTION = "Security analysis tool for Ethereum smart contracts"
URL = "https://github.com/ConsenSys/mythril"
AUTHOR = "ConsenSys Dilligence"
AUTHOR_MAIL = None
REQUIRES_PYTHON = ">=3.6.0"
here = os.path.abspath(os.path.dirname(__file__))

# What packages are required for this module to be executed?
def get_requirements():
    """
    Return requirements as list.
    Handles cases where git links are used.

    """
    with open(os.path.join(here, "requirements.txt")) as f:
        packages = []
        for line in f:
            line = line.strip()
            # let's also ignore empty lines and comments
            if not line or line.startswith("#"):
                continue
            if "https://" not in line:
                packages.append(line)
                continue

            rest, package_name = line.split("#egg=")[0], line.split("#egg=")[1]
            if "-e" in rest:
                rest = rest.split("-e")[1]
            package_name = package_name + "@" + rest
            packages.append(package_name)
    return packages


REQUIRED = get_requirements()

TESTS_REQUIRE = ["mypy==0.782", "pytest>=3.6.0", "pytest_mock", "pytest-cov"]

# What packages are optional?
EXTRAS = {}

# If version is set to None then it will be fetched from __version__.py
VERSION = None


# Import the README and use it as the long-description.
# Note: this will only work if 'README.md' is present in your MANIFEST.in file!
try:
    with io.open(os.path.join(here, "README.md"), encoding="utf-8") as f:
        long_description = "\n" + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION


# Load the package's __version__.py module as a dictionary.
about = {}
if not VERSION:
    project_slug = NAME.lower().replace("-", "_").replace(" ", "_")
    with open(os.path.join(here, project_slug, "__version__.py")) as f:
        exec(f.read(), about)
else:
    about["__version__"] = VERSION


# Package version (vX.Y.Z). It must match git tag being used for CircleCI
# deployment; otherwise the build will failed.
class VerifyVersionCommand(_install):
    """Custom command to verify that the git tag matches our version."""

    description = "verify that the git tag matches our version"

    def run(self):
        """"""
        tag = os.getenv("CIRCLE_TAG")

        if tag != about["__version__"]:
            info = "Git tag: {0} does not match the version of this app: {1}".format(
                tag, about["__version__"]
            )
            sys.exit(info)


setup(
    name=NAME,
    version=about["__version__"][1:],
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/markdown",  # requires twine and recent setuptools
    url=URL,
    author=AUTHOR,
    author_mail=AUTHOR_MAIL,
    license="MIT",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "Topic :: Software Development :: Disassemblers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    keywords="hacking disassembler security ethereum",
    packages=find_packages(exclude=["contrib", "docs", "tests"]),
    install_requires=REQUIRED,
    tests_require=TESTS_REQUIRE,
    python_requires=REQUIRES_PYTHON,
    extras_require=EXTRAS,
    package_data={"mythril.analysis.templates": ["*"], "mythril.support.assets": ["*"]},
    include_package_data=True,
    entry_points={"console_scripts": ["myth=mythril.interfaces.cli:main"]},
    cmdclass={"install": InstallCommand, "verify": VerifyVersionCommand},
)



"""This module contains analysis module helpers to solve path constraints."""

from typing import Dict, List, Tuple, Union, Any

import z3
import logging

from z3 import FuncInterp

from mythril.exceptions import UnsatError
from mythril.laser.ethereum.function_managers import (
    keccak_function_manager,
)

from mythril.laser.ethereum.state.constraints import Constraints
from mythril.laser.ethereum.state.global_state import GlobalState
from mythril.laser.ethereum.transaction import BaseTransaction
from mythril.laser.ethereum.transaction.transaction_models import (
    ContractCreationTransaction,
)
from mythril.laser.smt import UGE, symbol_factory
from mythril.support.model import get_model

log = logging.getLogger(__name__)
z3.set_option(
    max_args=10000000, max_lines=1000000, max_depth=10000000, max_visited=1000000
)


def pretty_print_model(model):
    """Pretty prints a z3 model

    :param model:
    :return:
    """
    ret = ""

    for d in model.decls():
        if type(model[d]) == FuncInterp:
            condition = model[d].as_list()
            ret += "%s: %s\n" % (d.name(), condition)
            continue

        try:
            condition = "0x%x" % model[d].as_long()
        except:
            condition = str(z3.simplify(model[d]))

        ret += "%s: %s\n" % (d.name(), condition)

    return ret


def get_transaction_sequence(
    global_state: GlobalState, constraints: Constraints
) -> Dict[str, Any]:
    """Generate concrete transaction sequence.
    Note: This function only considers the constraints in constraint argument,
    which in some cases is expected to differ from global_state's constraints

    :param global_state: GlobalState to generate transaction sequence for
    :param constraints: list of constraints used to generate transaction sequence
    """
    transaction_sequence = global_state.world_state.transaction_sequence
    concrete_transactions = []
    tx_constraints, minimize = _set_minimisation_constraints(
        transaction_sequence, constraints.copy(), [], 5000, global_state.world_state
    )

    try:
        model = get_model(tx_constraints, minimize=minimize)
    except UnsatError:
        raise UnsatError

    if isinstance(transaction_sequence[0], ContractCreationTransaction):
        initial_world_state = transaction_sequence[0].prev_world_state
    else:
        initial_world_state = transaction_sequence[0].world_state

    initial_accounts = initial_world_state.accounts

    for transaction in transaction_sequence:
        concrete_transaction = _get_concrete_transaction(model, transaction)
        concrete_transactions.append(concrete_transaction)

    min_price_dict: Dict[str, int] = {}
    for address in initial_accounts.keys():
        min_price_dict[address] = model.eval(
            initial_world_state.starting_balances[
                symbol_factory.BitVecVal(address, 256)
            ].raw,
            model_completion=True,
        ).as_long()

    concrete_initial_state = _get_concrete_state(initial_accounts, min_price_dict)
    if isinstance(transaction_sequence[0], ContractCreationTransaction):
        code = transaction_sequence[0].code
        _replace_with_actual_sha(concrete_transactions, model, code)
    else:
        _replace_with_actual_sha(concrete_transactions, model)
    _add_calldata_placeholder(concrete_transactions, transaction_sequence)
    steps = {"initialState": concrete_initial_state, "steps": concrete_transactions}

    return steps


def _add_calldata_placeholder(
    concrete_transactions: List[Dict[str, str]],
    transaction_sequence: List[BaseTransaction],
):
    """
    Adds a calldata placeholder into the concrete transactions
    :param concrete_transactions:
    :param transaction_sequence:
    :return:
    """
    for tx in concrete_transactions:
        tx["calldata"] = tx["input"]
    if not isinstance(transaction_sequence[0], ContractCreationTransaction):
        return

    if type(transaction_sequence[0].code.bytecode) == tuple:
        code_len = len(transaction_sequence[0].code.bytecode) * 2
    else:
        code_len = len(transaction_sequence[0].code.bytecode)
    concrete_transactions[0]["calldata"] = concrete_transactions[0]["input"][
        code_len + 2 :
    ]


def _replace_with_actual_sha(
    concrete_transactions: List[Dict[str, str]], model: z3.Model, code=None
):
    concrete_hashes = keccak_function_manager.get_concrete_hash_data(model)
    for tx in concrete_transactions:
        if keccak_function_manager.hash_matcher not in tx["input"]:
            continue
        if code is not None and code.bytecode in tx["input"]:
            s_index = len(code.bytecode) + 2
        else:
            s_index = 10
        for i in range(s_index, len(tx["input"])):
            data_slice = tx["input"][i : i + 64]
            if (
                keccak_function_manager.hash_matcher not in data_slice
                or len(data_slice) != 64
            ):
                continue
            find_input = symbol_factory.BitVecVal(int(data_slice, 16), 256)
            input_ = None
            for size in concrete_hashes:
                _, inverse = keccak_function_manager.store_function[size]
                if find_input.value not in concrete_hashes[size]:
                    continue
                input_ = symbol_factory.BitVecVal(
                    model.eval(inverse(find_input).raw).as_long(), size
                )

            if input_ is None:
                continue
            keccak = keccak_function_manager.find_concrete_keccak(input_)
            hex_keccak = hex(keccak.value)[2:]
            if len(hex_keccak) != 64:
                hex_keccak = "0" * (64 - len(hex_keccak)) + hex_keccak
            tx["input"] = tx["input"][:s_index] + tx["input"][s_index:].replace(
                tx["input"][i : 64 + i], hex_keccak
            )


def _get_concrete_state(
    initial_accounts: Dict, min_price_dict: Dict[str, int]
) -> Dict[str, Dict]:
    """Gets a concrete state"""
    accounts = {}
    for address, account in initial_accounts.items():
        # Skip empty default account

        data: Dict[str, Union[int, str]] = {}
        data["nonce"] = account.nonce
        data["code"] = account.serialised_code()
        data["storage"] = str(account.storage)
        data["balance"] = hex(min_price_dict.get(address, 0))
        accounts[hex(address)] = data
    return {"accounts": accounts}


def _get_concrete_transaction(model: z3.Model, transaction: BaseTransaction):
    """Gets a concrete transaction from a transaction and z3 model"""
    # Get concrete values from transaction
    address = hex(transaction.callee_account.address.value)
    value = model.eval(transaction.call_value.raw, model_completion=True).as_long()
    caller = "0x" + (
        "%x" % model.eval(transaction.caller.raw, model_completion=True).as_long()
    ).zfill(40)

    input_ = ""
    if isinstance(transaction, ContractCreationTransaction):
        address = ""
        input_ += transaction.code.bytecode

    input_ += "".join(
        [
            hex(b)[2:] if len(hex(b)) % 2 == 0 else "0" + hex(b)[2:]
            for b in transaction.call_data.concrete(model)
        ]
    )

    # Create concrete transaction dict
    concrete_transaction = dict()  # type: Dict[str, str]
    concrete_transaction["input"] = "0x" + input_
    concrete_transaction["value"] = "0x%x" % value
    # Fixme: base origin assignment on origin symbol
    concrete_transaction["origin"] = caller
    concrete_transaction["address"] = "%s" % address

    return concrete_transaction


def _set_minimisation_constraints(
    transaction_sequence, constraints, minimize, max_size, world_state
) -> Tuple[Constraints, tuple]:
    """Set constraints that minimise key transaction values

    Constraints generated:
    - Upper bound on calldata size
    - Minimisation of call value's and calldata sizes

    :param transaction_sequence: Transaction for which the constraints should be applied
    :param constraints: The constraints array which should contain any added constraints
    :param minimize: The minimisation array which should contain any variables that should be minimised
    :param max_size: The max size of the calldata array
    :return: updated constraints, minimize
    """
    for transaction in transaction_sequence:
        # Set upper bound on calldata size
        max_calldata_size = symbol_factory.BitVecVal(max_size, 256)
        constraints.append(UGE(max_calldata_size, transaction.call_data.calldatasize))

        # Minimize
        minimize.append(transaction.call_data.calldatasize)
        minimize.append(transaction.call_value)
        constraints.append(
            UGE(
                symbol_factory.BitVecVal(1000000000000000000000, 256),
                world_state.starting_balances[transaction.caller],
            )
        )

    for account in world_state.accounts.values():
        # Lazy way to prevent overflows and to ensure "reasonable" balances
        # Each account starts with less than 100 ETH
        constraints.append(
            UGE(
                symbol_factory.BitVecVal(100000000000000000000, 256),
                world_state.starting_balances[account.address],
            )
        )

    return constraints, tuple(minimize)



"""This module implements the main symbolic execution engine."""
import logging
from collections import defaultdict
from copy import copy
from datetime import datetime, timedelta
import random
from typing import Callable, Dict, DefaultDict, List, Tuple, Optional

from mythril.support.opcodes import OPCODES
from mythril.analysis.potential_issues import check_potential_issues
from mythril.laser.execution_info import ExecutionInfo
from mythril.laser.ethereum.cfg import NodeFlags, Node, Edge, JumpType
from mythril.laser.ethereum.evm_exceptions import StackUnderflowException, VmException
from mythril.laser.ethereum.instructions import Instruction
from mythril.laser.ethereum.instruction_data import get_required_stack_elements
from mythril.laser.plugin.signals import PluginSkipWorldState, PluginSkipState
from mythril.laser.ethereum.state.global_state import GlobalState
from mythril.laser.ethereum.state.world_state import WorldState
from mythril.laser.ethereum.strategy.basic import DepthFirstSearchStrategy
from mythril.laser.ethereum.strategy.constraint_strategy import DelayConstraintStrategy
from abc import ABCMeta
from mythril.laser.ethereum.time_handler import time_handler

from mythril.laser.ethereum.transaction import (
    ContractCreationTransaction,
    TransactionEndSignal,
    TransactionStartSignal,
    execute_contract_creation,
    execute_message_call,
)
from mythril.laser.smt import And, symbol_factory, simplify
from mythril.support.support_args import args

log = logging.getLogger(__name__)


class SVMError(Exception):
    """An exception denoting an unexpected state in symbolic execution."""

    pass


class LaserEVM:
    """The LASER EVM.

    Just as Mithril had to be mined at great efforts to provide the
    Dwarves with their exceptional armour, LASER stands at the heart of
    Mythril, digging deep in the depths of call graphs, unearthing the
    most precious symbolic call data, that is then hand-forged into
    beautiful and strong security issues by the experienced smiths we
    call detection modules. It is truly a magnificent symbiosis.
    """

    def __init__(
        self,
        dynamic_loader=None,
        max_depth=float("inf"),
        execution_timeout=60,
        create_timeout=10,
        strategy=DepthFirstSearchStrategy,
        transaction_count=2,
        requires_statespace=True,
        iprof=None,
        use_reachability_check=True,
        beam_width=None,
    ) -> None:
        """
        Initializes the laser evm object

        :param dynamic_loader: Loads data from chain
        :param max_depth: Maximum execution depth this vm should execute
        :param execution_timeout: Time to take for execution
        :param create_timeout: Time to take for contract creation
        :param strategy: Execution search strategy
        :param transaction_count: The amount of transactions to execute
        :param requires_statespace: Variable indicating whether the statespace should be recorded
        :param iprof: Instruction Profiler
        """
        self.execution_info: List[ExecutionInfo] = []

        self.open_states: List[WorldState] = []
        self.total_states = 0
        self.dynamic_loader = dynamic_loader
        self.use_reachability_check = use_reachability_check

        self.work_list: List[GlobalState] = []
        self.strategy = strategy(self.work_list, max_depth, beam_width=beam_width)
        self.max_depth = max_depth
        self.transaction_count = transaction_count

        self.execution_timeout = execution_timeout or 0
        self.create_timeout = create_timeout or 0

        self.requires_statespace = requires_statespace
        if self.requires_statespace:
            self.nodes: Dict[int, Node] = {}
            self.edges: List[Edge] = []

        self.time: datetime = None
        self.executed_transactions: bool = False

        self.pre_hooks: DefaultDict[str, List[Callable]] = defaultdict(list)
        self.post_hooks: DefaultDict[str, List[Callable]] = defaultdict(list)

        self._add_world_state_hooks: List[Callable] = []
        self._execute_state_hooks: List[Callable] = []

        self._start_exec_trans_hooks: List[Callable] = []
        self._stop_exec_trans_hooks: List[Callable] = []

        self._start_sym_trans_hooks: List[Callable] = []
        self._stop_sym_trans_hooks: List[Callable] = []

        self._start_sym_exec_hooks: List[Callable] = []
        self._stop_sym_exec_hooks: List[Callable] = []

        self._start_exec_hooks: List[Callable] = []
        self._stop_exec_hooks: List[Callable] = []

        self._transaction_end_hooks: List[Callable] = []

        self.iprof = iprof
        self.instr_pre_hook: Dict[str, List[Callable]] = {}
        self.instr_post_hook: Dict[str, List[Callable]] = {}
        for op in OPCODES:
            self.instr_pre_hook[op] = []
            self.instr_post_hook[op] = []
        self.hook_type_map = {
            "start_execute_transactions": self._start_exec_trans_hooks,
            "stop_execute_transactions": self._stop_exec_trans_hooks,
            "add_world_state": self._add_world_state_hooks,
            "execute_state": self._execute_state_hooks,
            "start_sym_exec": self._start_sym_exec_hooks,
            "stop_sym_exec": self._stop_sym_exec_hooks,
            "start_sym_trans": self._start_sym_trans_hooks,
            "stop_sym_trans": self._stop_sym_trans_hooks,
            "start_exec": self._start_exec_hooks,
            "stop_exec": self._stop_exec_hooks,
            "transaction_end": self._transaction_end_hooks,
        }
        log.info("LASER EVM initialized with dynamic loader: " + str(dynamic_loader))

    def extend_strategy(self, extension: ABCMeta, **kwargs) -> None:
        self.strategy = extension(self.strategy, **kwargs)

    def sym_exec(
        self,
        world_state: WorldState = None,
        target_address: int = None,
        creation_code: str = None,
        contract_name: str = None,
    ) -> None:
        """Starts symbolic execution
        There are two modes of execution.
        Either we analyze a preconfigured configuration, in which case the world_state and target_address variables
        must be supplied.
        Or we execute the creation code of a contract, in which case the creation code and desired name of that
        contract should be provided.

        :param world_state The world state configuration from which to perform analysis
        :param target_address The address of the contract account in the world state which analysis should target
        :param creation_code The creation code to create the target contract in the symbolic environment
        :param contract_name The name that the created account should be associated with
        """
        pre_configuration_mode = target_address is not None
        scratch_mode = creation_code is not None and contract_name is not None
        if pre_configuration_mode == scratch_mode:
            raise ValueError("Symbolic execution started with invalid parameters")

        log.debug("Starting LASER execution")
        for hook in self._start_sym_exec_hooks:
            hook()

        time_handler.start_execution(self.execution_timeout)
        self.time = datetime.now()

        if pre_configuration_mode:
            self.open_states = [world_state]
            log.info("Starting message call transaction to {}".format(target_address))
            self.execute_transactions(symbol_factory.BitVecVal(target_address, 256))

        elif scratch_mode:
            log.info("Starting contract creation transaction")

            created_account = execute_contract_creation(
                self, creation_code, contract_name, world_state=world_state
            )
            log.info(
                "Finished contract creation, found {} open states".format(
                    len(self.open_states)
                )
            )

            if len(self.open_states) == 0:
                log.warning(
                    "No contract was created during the execution of contract creation "
                    "Increase the resources for creation execution (--max-depth or --create-timeout) "
                    "Check whether the bytecode is indeed the creation code, otherwise use the --bin-runtime flag"
                )

            self.execute_transactions(created_account.address)

        log.info("Finished symbolic execution")
        if self.requires_statespace:
            log.info(
                "%d nodes, %d edges, %d total states",
                len(self.nodes),
                len(self.edges),
                self.total_states,
            )

        for hook in self._stop_sym_exec_hooks:
            hook()

    def execute_transactions(self, address) -> None:
        """This function helps runs plugins that can order transactions.
        Such plugins should set self.executed_transactions as True after its execution

        :param address: Address of the contract
        :return: None
        """
        for hook in self._start_exec_trans_hooks:
            hook()

        if self.executed_transactions is False:
            self._execute_transactions(address)

        for hook in self._stop_exec_trans_hooks:
            hook()

    def _execute_transactions(self, address):
        """This function executes multiple transactions on the address

        :param address: Address of the contract
        :return:
        """
        self.time = datetime.now()

        for i in range(self.transaction_count):
            if len(self.open_states) == 0:
                break
            old_states_count = len(self.open_states)

            if self.use_reachability_check:
                if isinstance(self.strategy, DelayConstraintStrategy):
                    open_states = []
                    for state in self.open_states:
                        c_val = self.strategy.model_cache.check_quick_sat(
                            simplify(And(*state.world_state.constraints)).raw
                        )
                        if c_val:
                            open_states.append(self.open_states)
                        else:
                            self.strategy.pending_worklist.append(state)

                else:
                    self.open_states = [
                        state
                        for state in self.open_states
                        if state.constraints.is_possible()
                    ]
                    prune_count = old_states_count - len(self.open_states)
                if prune_count:
                    log.info("Pruned {} unreachable states".format(prune_count))
            log.info(
                "Starting message call transaction, iteration: {}, {} initial states".format(
                    i, len(self.open_states)
                )
            )
            func_hashes = (
                args.transaction_sequences[i] if args.transaction_sequences else None
            )

            if func_hashes:
                for itr, func_hash in enumerate(func_hashes):
                    if func_hash in (-1, -2):
                        func_hashes[itr] = func_hash
                    else:
                        func_hashes[itr] = bytes.fromhex(hex(func_hash)[2:].zfill(8))

            for hook in self._start_sym_trans_hooks:
                hook()

            execute_message_call(self, address, func_hashes=func_hashes)

            for hook in self._stop_sym_trans_hooks:
                hook()

        self.executed_transactions = True

    def _check_create_termination(self) -> bool:
        if len(self.open_states) != 0:
            return (
                self.create_timeout > 0
                and self.time + timedelta(seconds=self.create_timeout) <= datetime.now()
            )
        return self._check_execution_termination()

    def _check_execution_termination(self) -> bool:
        return (
            self.execution_timeout > 0
            and self.time + timedelta(seconds=self.execution_timeout) <= datetime.now()
        )

    def exec(self, create=False, track_gas=False) -> Optional[List[GlobalState]]:
        """

        :param create:
        :param track_gas:
        :return:
        """
        final_states = []  # type: List[GlobalState]
        for hook in self._start_exec_hooks:
            hook()

        for global_state in self.strategy:

            if create and self._check_create_termination():
                log.debug("Hit create timeout, returning.")
                return final_states + [global_state] if track_gas else None

            if not create and self._check_execution_termination():
                log.debug("Hit execution timeout, returning.")
                return final_states + [global_state] if track_gas else None
            try:
                new_states, op_code = self.execute_state(global_state)
            except NotImplementedError:
                log.debug("Encountered unimplemented instruction")
                continue

            if self.strategy.run_check() and (
                len(new_states) > 1 and random.uniform(0, 1) < args.pruning_factor
            ):
                new_states = [
                    state
                    for state in new_states
                    if state.world_state.constraints.is_possible()
                ]
            self.manage_cfg(op_code, new_states)  # TODO: What about op_code is None?
            if new_states:
                self.work_list += new_states
            elif track_gas:
                final_states.append(global_state)
            self.total_states += len(new_states)

        for hook in self._stop_exec_hooks:
            hook()

        return final_states if track_gas else None

    def _add_world_state(self, global_state: GlobalState):
        """Stores the world_state of the passed global state in the open states"""

        for hook in self._add_world_state_hooks:
            try:
                hook(global_state)
            except PluginSkipWorldState:
                return

        self.open_states.append(global_state.world_state)

    def handle_vm_exception(
        self, global_state: GlobalState, op_code: str, error_msg: str
    ) -> List[GlobalState]:
        _, return_global_state = global_state.transaction_stack.pop()

        if return_global_state is None:
            # In this case we don't put an unmodified world state in the open_states list Since in the case of an
            #  exceptional halt all changes should be discarded, and this world state would not provide us with a
            #  previously unseen world state
            log.debug("Encountered a VmException, ending path: `{}`".format(error_msg))
            new_global_states = []  # type: List[GlobalState]
        else:
            # First execute the post hook for the transaction ending instruction
            self._execute_post_hook(op_code, [global_state])
            new_global_states = self._end_message_call(
                return_global_state, global_state, revert_changes=True, return_data=None
            )
        return new_global_states

    def execute_state(
        self, global_state: GlobalState
    ) -> Tuple[List[GlobalState], Optional[str]]:
        """Execute a single instruction in global_state.

        :param global_state:
        :return: A list of successor states.
        """
        # Execute hooks
        try:
            for hook in self._execute_state_hooks:
                hook(global_state)
        except PluginSkipState:
            return [], None

        instructions = global_state.environment.code.instruction_list
        try:
            op_code = instructions[global_state.mstate.pc]["opcode"]
        except IndexError:
            self._add_world_state(global_state)
            return [], None

        if len(global_state.mstate.stack) < get_required_stack_elements(op_code):
            error_msg = (
                "Stack Underflow Exception due to insufficient "
                "stack elements for the address {}".format(
                    instructions[global_state.mstate.pc]["address"]
                )
            )
            new_global_states = self.handle_vm_exception(
                global_state, op_code, error_msg
            )
            self._execute_post_hook(op_code, new_global_states)
            return new_global_states, op_code

        try:
            self._execute_pre_hook(op_code, global_state)
        except PluginSkipState:
            return [], None

        try:
            new_global_states = Instruction(
                op_code,
                self.dynamic_loader,
                pre_hooks=self.instr_pre_hook[op_code],
                post_hooks=self.instr_post_hook[op_code],
            ).evaluate(global_state)

        except VmException as e:
            for hook in self._transaction_end_hooks:
                hook(
                    global_state,
                    global_state.current_transaction,
                    None,
                    False,
                )
            new_global_states = self.handle_vm_exception(global_state, op_code, str(e))

        except TransactionStartSignal as start_signal:
            # Setup new global state
            new_global_state = start_signal.transaction.initial_global_state()

            new_global_state.transaction_stack = copy(
                global_state.transaction_stack
            ) + [(start_signal.transaction, global_state)]
            new_global_state.node = global_state.node
            new_global_state.world_state.constraints = (
                start_signal.global_state.world_state.constraints
            )

            log.debug("Starting new transaction %s", start_signal.transaction)

            return [new_global_state], op_code

        except TransactionEndSignal as end_signal:
            (
                transaction,
                return_global_state,
            ) = end_signal.global_state.transaction_stack[-1]

            log.debug("Ending transaction %s.", transaction)

            for hook in self._transaction_end_hooks:
                hook(
                    end_signal.global_state,
                    transaction,
                    return_global_state,
                    end_signal.revert,
                )

            if return_global_state is None:
                if (
                    not isinstance(transaction, ContractCreationTransaction)
                    or transaction.return_data
                ) and not end_signal.revert:
                    check_potential_issues(global_state)
                    end_signal.global_state.world_state.node = global_state.node
                    self._add_world_state(end_signal.global_state)

                new_global_states = []
            else:

                # First execute the post hook for the transaction ending instruction
                self._execute_post_hook(op_code, [end_signal.global_state])

                # Propagate annotations
                new_annotations = [
                    annotation
                    for annotation in global_state.annotations
                    if annotation.persist_over_calls
                ]
                return_global_state.add_annotations(new_annotations)

                new_global_states = self._end_message_call(
                    copy(return_global_state),
                    global_state,
                    revert_changes=False or end_signal.revert,
                    return_data=transaction.return_data,
                )

        self._execute_post_hook(op_code, new_global_states)

        return new_global_states, op_code

    def _end_message_call(
        self,
        return_global_state: GlobalState,
        global_state: GlobalState,
        revert_changes=False,
        return_data=None,
    ) -> List[GlobalState]:
        """

        :param return_global_state:
        :param global_state:
        :param revert_changes:
        :param return_data:
        :return:
        """

        return_global_state.world_state.constraints += (
            global_state.world_state.constraints
        )
        # Resume execution of the transaction initializing instruction
        op_code = return_global_state.environment.code.instruction_list[
            return_global_state.mstate.pc
        ]["opcode"]

        # Set execution result in the return_state
        return_global_state.last_return_data = return_data
        if not revert_changes:
            return_global_state.world_state = copy(global_state.world_state)
            return_global_state.environment.active_account = global_state.accounts[
                return_global_state.environment.active_account.address.value
            ]
            if isinstance(
                global_state.current_transaction, ContractCreationTransaction
            ):
                return_global_state.mstate.min_gas_used += (
                    global_state.mstate.min_gas_used
                )
                return_global_state.mstate.max_gas_used += (
                    global_state.mstate.max_gas_used
                )
        try:
            # Execute the post instruction handler
            new_global_states = Instruction(
                op_code,
                self.dynamic_loader,
                pre_hooks=self.instr_pre_hook[op_code],
                post_hooks=self.instr_post_hook[op_code],
            ).evaluate(return_global_state, True)
        except VmException:
            new_global_states = []
        # In order to get a nice call graph we need to set the nodes here
        for state in new_global_states:
            state.node = global_state.node

        return new_global_states

    def manage_cfg(self, opcode: str, new_states: List[GlobalState]) -> None:
        """

        :param opcode:
        :param new_states:
        """
        if opcode == "JUMP":
            assert len(new_states) <= 1
            for state in new_states:
                self._new_node_state(state)
        elif opcode == "JUMPI":
            assert len(new_states) <= 2
            for state in new_states:
                self._new_node_state(
                    state, JumpType.CONDITIONAL, state.world_state.constraints[-1]
                )
        elif opcode in ("SLOAD", "SSTORE") and len(new_states) > 1:
            for state in new_states:
                self._new_node_state(
                    state, JumpType.CONDITIONAL, state.world_state.constraints[-1]
                )
        elif opcode == "RETURN":
            for state in new_states:
                self._new_node_state(state, JumpType.RETURN)

        for state in new_states:
            state.node.states.append(state)

    def _new_node_state(
        self, state: GlobalState, edge_type=JumpType.UNCONDITIONAL, condition=None
    ) -> None:
        """

        :param state:
        :param edge_type:
        :param condition:
        """
        try:
            address = state.environment.code.instruction_list[state.mstate.pc][
                "address"
            ]
        except IndexError:
            return
        new_node = Node(state.environment.active_account.contract_name)
        old_node = state.node
        state.node = new_node
        new_node.constraints = state.world_state.constraints
        if self.requires_statespace:
            self.nodes[new_node.uid] = new_node
            self.edges.append(
                Edge(
                    old_node.uid, new_node.uid, edge_type=edge_type, condition=condition
                )
            )

        if edge_type == JumpType.RETURN:
            new_node.flags |= NodeFlags.CALL_RETURN
        elif edge_type == JumpType.CALL:
            try:
                if "retval" in str(state.mstate.stack[-1]):
                    new_node.flags |= NodeFlags.CALL_RETURN
                else:
                    new_node.flags |= NodeFlags.FUNC_ENTRY
            except StackUnderflowException:
                new_node.flags |= NodeFlags.FUNC_ENTRY

        environment = state.environment
        disassembly = environment.code

        # Only set when encountering new JUMPI
        if edge_type == JumpType.CONDITIONAL:
            if isinstance(
                state.world_state.transaction_sequence[-1], ContractCreationTransaction
            ):
                environment.active_function_name = "constructor"
            elif address in disassembly.address_to_function_name:
                # Enter a new function
                environment.active_function_name = disassembly.address_to_function_name[
                    address
                ]
                new_node.flags |= NodeFlags.FUNC_ENTRY

                log.debug(
                    "- Entering function "
                    + environment.active_account.contract_name
                    + ":"
                    + new_node.function_name
                )
            elif address == 0:
                environment.active_function_name = "fallback"

        new_node.function_name = environment.active_function_name

    def register_hooks(self, hook_type: str, hook_dict: Dict[str, List[Callable]]):
        """

        :param hook_type:
        :param hook_dict:
        """
        if hook_type == "pre":
            entrypoint = self.pre_hooks
        elif hook_type == "post":
            entrypoint = self.post_hooks
        else:
            raise ValueError(
                "Invalid hook type %s. Must be one of {pre, post}", hook_type
            )

        for op_code, funcs in hook_dict.items():
            entrypoint[op_code].extend(funcs)

    def register_laser_hooks(self, hook_type: str, hook: Callable):
        """registers the hook with this Laser VM"""

        if hook_type in self.hook_type_map:
            self.hook_type_map[hook_type].append(hook)
        else:
            raise ValueError(f"Invalid hook type {hook_type}")

    def register_instr_hooks(self, hook_type: str, opcode: str, hook: Callable):
        """Registers instructions hooks from plugins"""
        if hook_type == "pre":
            if opcode is None:
                for op in OPCODES:
                    self.instr_pre_hook[op].append(hook(op))
            else:
                self.instr_pre_hook[opcode].append(hook)
        else:
            if opcode is None:
                for op in OPCODES:
                    self.instr_post_hook[op].append(hook(op))
            else:
                self.instr_post_hook[opcode].append(hook)

    def instr_hook(self, hook_type, opcode) -> Callable:
        """Registers the annoted function with register_instr_hooks

        :param hook_type: Type of hook pre/post
        :param opcode: The opcode related to the function
        """

        def hook_decorator(func: Callable):
            """Hook decorator generated by laser_hook

            :param func: Decorated function
            """
            self.register_instr_hooks(hook_type, opcode, func)

        return hook_decorator

    def laser_hook(self, hook_type: str) -> Callable:
        """Registers the annotated function with register_laser_hooks

        :param hook_type:
        :return: hook decorator
        """

        def hook_decorator(func: Callable):
            """Hook decorator generated by laser_hook

            :param func: Decorated function
            """
            self.register_laser_hooks(hook_type, func)
            return func

        return hook_decorator

    def _execute_pre_hook(self, op_code: str, global_state: GlobalState) -> None:
        """

        :param op_code:
        :param global_state:
        :return:
        """
        if op_code not in self.pre_hooks.keys():
            return
        for hook in self.pre_hooks[op_code]:
            hook(global_state)

    def _execute_post_hook(
        self, op_code: str, global_states: List[GlobalState]
    ) -> None:
        """

        :param op_code:
        :param global_states:
        :return:
        """
        if op_code not in self.post_hooks.keys():
            return

        for hook in self.post_hooks[op_code]:
            for global_state in global_states:
                try:
                    hook(global_state)
                except PluginSkipState:
                    global_states.remove(global_state)

    def pre_hook(self, op_code: str) -> Callable:
        """

        :param op_code:
        :return:
        """

        def hook_decorator(func: Callable):
            """

            :param func:
            :return:
            """
            if op_code not in self.pre_hooks.keys():
                self.pre_hooks[op_code] = []
            self.pre_hooks[op_code].append(func)
            return func

        return hook_decorator

    def post_hook(self, op_code: str) -> Callable:
        """

        :param op_code:
        :return:
        """

        def hook_decorator(func: Callable):
            """

            :param func:
            :return:
            """
            if op_code not in self.post_hooks.keys():
                self.post_hooks[op_code] = []
            self.post_hooks[op_code].append(func)
            return func

        return hook_decorator



from eth._utils.numeric import ceil32
from typing import Tuple
from mythril.support.opcodes import OPCODES, STACK, GAS
from eth.constants import (
    GAS_ECRECOVER,
    GAS_SHA256WORD,
    GAS_SHA256,
    GAS_RIPEMD160,
    GAS_RIPEMD160WORD,
    GAS_IDENTITY,
    GAS_IDENTITYWORD,
    GAS_SHA3WORD,
    GAS_SHA3,
)


def calculate_sha3_gas(length: int):
    """

    :param length:
    :return:
    """
    gas_val = GAS_SHA3 + GAS_SHA3WORD * (ceil32(length) // 32)
    return gas_val, gas_val


def calculate_native_gas(size: int, contract: str):
    """

    :param size:
    :param contract:
    :return:
    """
    gas_value = 0
    word_num = ceil32(size) // 32
    if contract == "ecrecover":
        gas_value = GAS_ECRECOVER
    elif contract == "sha256":
        gas_value = GAS_SHA256 + word_num * GAS_SHA256WORD
    elif contract == "ripemd160":
        gas_value = GAS_RIPEMD160 + word_num * GAS_RIPEMD160WORD
    elif contract == "identity":
        gas_value = GAS_IDENTITY + word_num * GAS_IDENTITYWORD
    else:
        # TODO: Add gas for other precompiles, computation should be shifted to natives.py
        #  as some data isn't available here
        pass
    return gas_value, gas_value


def get_opcode_gas(opcode: str) -> Tuple[int, int]:
    return OPCODES[opcode][GAS]


def get_required_stack_elements(opcode: str) -> int:
    return OPCODES[opcode][STACK][0]



from mythril.laser.smt import symbol_factory
from mythril.disassembler.disassembly import Disassembly
from mythril.laser.ethereum.state.environment import Environment
from mythril.laser.ethereum.state.machine_state import MachineState
from mythril.laser.ethereum.state.global_state import GlobalState
from mythril.laser.ethereum.state.world_state import WorldState
from mythril.laser.ethereum.instructions import Instruction
from mythril.laser.ethereum.transaction.transaction_models import MessageCallTransaction


def test_extcodecopy():
    # Arrange
    new_world_state = WorldState()
    new_account = new_world_state.create_account(balance=10, address=101)
    new_account.code = Disassembly("60616240")
    ext_account = new_world_state.create_account(balance=1000, address=121)
    ext_account.code = Disassembly("6040404040")

    new_environment = Environment(new_account, None, None, None, None, None, None)
    state = GlobalState(
        new_world_state, new_environment, None, MachineState(gas_limit=8000000)
    )
    state.transaction_stack.append(
        (MessageCallTransaction(world_state=WorldState(), gas_limit=8000000), None)
    )

    state.mstate.stack = [3, 0, 0, 121]
    instruction = Instruction("extcodecopy", dynamic_loader=None)

    # Act
    new_state = instruction.evaluate(state)[0]
    # Assert
    assert new_state.mstate.memory[0:3] == [96, 64, 64]


def test_extcodecopy_fail():
    # Arrange
    new_world_state = WorldState()
    new_account = new_world_state.create_account(balance=10, address=101)
    new_account.code = Disassembly("60616240")
    new_environment = Environment(new_account, None, None, None, None, None, None)
    state = GlobalState(
        new_world_state, new_environment, None, MachineState(gas_limit=8000000)
    )
    state.transaction_stack.append(
        (MessageCallTransaction(world_state=WorldState(), gas_limit=8000000), None)
    )

    state.mstate.stack = [2, 2, 2, symbol_factory.BitVecSym("FAIL", 256)]
    instruction = Instruction("extcodecopy", dynamic_loader=None)

    # Act
    new_state = instruction.evaluate(state)[0]

    # Assert
    assert new_state.mstate.stack == []
    assert new_state.mstate.memory._memory == state.mstate.memory._memory



"""This module contains the detection code to find multiple sends occurring in
a single transaction."""
from copy import copy
from typing import cast, List
from mythril.analysis.issue_annotation import IssueAnnotation
from mythril.analysis.report import Issue
from mythril.analysis.solver import get_transaction_sequence, UnsatError
from mythril.analysis.swc_data import MULTIPLE_SENDS
from mythril.analysis.module.base import DetectionModule, EntryPoint
from mythril.laser.ethereum.state.annotation import StateAnnotation
from mythril.laser.ethereum.state.global_state import GlobalState
from mythril.laser.smt import And
import logging

log = logging.getLogger(__name__)


class MultipleSendsAnnotation(StateAnnotation):
    def __init__(self) -> None:
        self.call_offsets = []  # type: List[int]

    def __copy__(self):
        result = MultipleSendsAnnotation()
        result.call_offsets = copy(self.call_offsets)
        return result


class MultipleSends(DetectionModule):
    """This module checks for multiple sends in a single transaction."""

    name = "Multiple external calls in the same transaction"
    swc_id = MULTIPLE_SENDS
    description = "Check for multiple sends in a single transaction"
    entry_point = EntryPoint.CALLBACK
    pre_hooks = ["CALL", "DELEGATECALL", "STATICCALL", "CALLCODE", "RETURN", "STOP"]

    def _execute(self, state: GlobalState) -> None:
        return self._analyze_state(state)

    def _analyze_state(self, state: GlobalState):
        """
        :param state: the current state
        :return: returns the issues for that corresponding state
        """
        instruction = state.get_current_instruction()

        annotations = cast(
            List[MultipleSendsAnnotation],
            list(state.get_annotations(MultipleSendsAnnotation)),
        )
        if len(annotations) == 0:
            state.annotate(MultipleSendsAnnotation())
            annotations = cast(
                List[MultipleSendsAnnotation],
                list(state.get_annotations(MultipleSendsAnnotation)),
            )

        call_offsets = annotations[0].call_offsets

        if instruction["opcode"] in ["CALL", "DELEGATECALL", "STATICCALL", "CALLCODE"]:
            call_offsets.append(state.get_current_instruction()["address"])

        else:  # RETURN or STOP

            for offset in call_offsets[1:]:
                try:
                    transaction_sequence = get_transaction_sequence(
                        state, state.world_state.constraints
                    )
                except UnsatError:
                    continue
                description_tail = (
                    "This call is executed following another call within the same transaction. It is possible "
                    "that the call never gets executed if a prior call fails permanently. This might be caused "
                    "intentionally by a malicious callee. If possible, refactor the code such that each transaction "
                    "only executes one external call or "
                    "make sure that all callees can be trusted (i.e. they’re part of your own codebase)."
                )

                issue = Issue(
                    contract=state.environment.active_account.contract_name,
                    function_name=state.environment.active_function_name,
                    address=offset,
                    swc_id=MULTIPLE_SENDS,
                    bytecode=state.environment.code.bytecode,
                    title="Multiple Calls in a Single Transaction",
                    severity="Low",
                    description_head="Multiple calls are executed in the same transaction.",
                    description_tail=description_tail,
                    gas_used=(state.mstate.min_gas_used, state.mstate.max_gas_used),
                    transaction_sequence=transaction_sequence,
                )
                state.annotate(
                    IssueAnnotation(
                        conditions=[And(*state.world_state.constraints)],
                        issue=issue,
                        detector=self,
                    )
                )
                return [issue]

        return []


detector = MultipleSends()



from mythril.laser.ethereum.transaction.symbolic import (
    execute_message_call,
    execute_contract_creation,
)
from mythril.laser.ethereum.transaction import (
    MessageCallTransaction,
    ContractCreationTransaction,
)
from mythril.laser.ethereum.svm import LaserEVM
from mythril.laser.ethereum.state.account import Account
from mythril.laser.ethereum.state.world_state import WorldState
from mythril.laser.smt import symbol_factory

import unittest.mock as mock
from unittest.mock import MagicMock


def _is_message_call(_, transaction, transaction_sequences):
    assert isinstance(transaction, MessageCallTransaction)


def _is_contract_creation(_, transaction):
    assert isinstance(transaction, ContractCreationTransaction)


@mock.patch(
    "mythril.laser.ethereum.transaction.symbolic._setup_global_state_for_execution"
)
def test_execute_message_call(mocked_setup: MagicMock):
    # Arrange
    laser_evm = LaserEVM({})

    world_state = WorldState()
    world_state.put_account(Account("0x0"))

    laser_evm.open_states = [world_state]
    laser_evm.exec = MagicMock()

    mocked_setup.side_effect = _is_message_call

    # Act
    execute_message_call(laser_evm, symbol_factory.BitVecVal(0, 256))

    # Assert
    # laser_evm.exec.assert_called_once()
    assert laser_evm.exec.call_count == 1
    # mocked_setup.assert_called_once()
    assert mocked_setup.call_count == 1

    assert len(laser_evm.open_states) == 0


@mock.patch(
    "mythril.laser.ethereum.transaction.symbolic._setup_global_state_for_execution"
)
def test_execute_contract_creation(mocked_setup: MagicMock):
    # Arrange
    laser_evm = LaserEVM({})

    laser_evm.open_states = [WorldState(), WorldState()]
    laser_evm.exec = MagicMock()
    mocked_setup.side_effect = _is_contract_creation

    # Act
    execute_contract_creation(laser_evm, "606000")

    # Assert
    # mocked_setup.assert_called()
    assert mocked_setup.call_count >= 1
    # laser_evm.exec.assert_called_once()
    assert laser_evm.exec.call_count == 1
    assert len(laser_evm.open_states) == 0



"""This module declares classes to represent call data."""
from typing import List

from mythril.laser.smt import (
    BitVec,
)


class ReturnData:
    """Base returndata class."""

    def __init__(self, return_data: List[BitVec], return_data_size: BitVec) -> None:
        """

        :param tx_id:
        """
        self.return_data = return_data
        self.return_data_size = return_data_size

    @property
    def size(self) -> BitVec:
        """

        :return: Calldata size for this calldata object
        """
        return self.return_data_size

    def __getitem__(self, index):
        if index < self.size:
            return self.return_data[index]
        else:
            return 0



from mythril.analysis.potential_issues import (
    PotentialIssue,
    get_potential_issues_annotation,
)
from mythril.analysis.swc_data import REENTRANCY
from mythril.analysis.module.base import DetectionModule, EntryPoint
from mythril.laser.ethereum.state.constraints import Constraints
from mythril.laser.smt import symbol_factory, UGT, BitVec, Or
from mythril.laser.ethereum.state.global_state import GlobalState
from mythril.laser.ethereum.state.annotation import StateAnnotation
from mythril.analysis import solver
from mythril.exceptions import UnsatError
from typing import List, cast, Optional
from copy import copy

import logging

log = logging.getLogger(__name__)

DESCRIPTION = """

Check whether the account state is accesses after the execution of an external call
"""

CALL_LIST = ["CALL", "DELEGATECALL", "CALLCODE"]
STATE_READ_WRITE_LIST = ["SSTORE", "SLOAD", "CREATE", "CREATE2"]


class StateChangeCallsAnnotation(StateAnnotation):
    def __init__(self, call_state: GlobalState, user_defined_address: bool) -> None:
        self.call_state = call_state
        self.state_change_states = []  # type: List[GlobalState]
        self.user_defined_address = user_defined_address

    def __copy__(self):
        new_annotation = StateChangeCallsAnnotation(
            self.call_state, self.user_defined_address
        )
        new_annotation.state_change_states = self.state_change_states[:]
        return new_annotation

    def get_issue(
        self, global_state: GlobalState, detector: DetectionModule
    ) -> Optional[PotentialIssue]:

        if not self.state_change_states:
            return None
        constraints = Constraints()
        gas = self.call_state.mstate.stack[-1]
        to = self.call_state.mstate.stack[-2]
        constraints += [
            UGT(gas, symbol_factory.BitVecVal(2300, 256)),
            Or(
                to > symbol_factory.BitVecVal(16, 256),
                to == symbol_factory.BitVecVal(0, 256),
            ),
        ]
        if self.user_defined_address:
            constraints += [to == 0xDEADBEEFDEADBEEFDEADBEEFDEADBEEFDEADBEEF]

        try:
            solver.get_transaction_sequence(
                global_state, constraints + global_state.world_state.constraints
            )
        except UnsatError:
            return None

        severity = "Medium" if self.user_defined_address else "Low"
        address = global_state.get_current_instruction()["address"]
        logging.debug(
            "[EXTERNAL_CALLS] Detected state changes at addresses: {}".format(address)
        )
        read_or_write = "Write to"
        if global_state.get_current_instruction()["opcode"] == "SLOAD":
            read_or_write = "Read of"
        address_type = "user defined" if self.user_defined_address else "fixed"
        description_head = "{} persistent state following external call".format(
            read_or_write
        )
        description_tail = (
            "The contract account state is accessed after an external call to a {} address. "
            "To prevent reentrancy issues, consider accessing the state only before the call, especially if the callee is untrusted. "
            "Alternatively, a reentrancy lock can be used to prevent "
            "untrusted callees from re-entering the contract in an intermediate state.".format(
                address_type
            )
        )

        return PotentialIssue(
            contract=global_state.environment.active_account.contract_name,
            function_name=global_state.environment.active_function_name,
            address=address,
            title="State access after external call",
            severity=severity,
            description_head=description_head,
            description_tail=description_tail,
            swc_id=REENTRANCY,
            bytecode=global_state.environment.code.bytecode,
            constraints=constraints,
            detector=detector,
        )


class StateChangeAfterCall(DetectionModule):
    """This module searches for state change after low level calls (e.g. call.value()) that
    forward gas to the callee."""

    name = "State change after an external call"
    swc_id = REENTRANCY
    description = DESCRIPTION
    entry_point = EntryPoint.CALLBACK
    pre_hooks = CALL_LIST + STATE_READ_WRITE_LIST

    def _execute(self, state: GlobalState) -> None:
        issues = self._analyze_state(state)

        annotation = get_potential_issues_annotation(state)
        annotation.potential_issues.extend(issues)

    @staticmethod
    def _add_external_call(global_state: GlobalState) -> None:
        gas = global_state.mstate.stack[-1]
        to = global_state.mstate.stack[-2]
        try:
            constraints = copy(global_state.world_state.constraints)
            solver.get_model(
                constraints
                + [
                    UGT(gas, symbol_factory.BitVecVal(2300, 256)),
                    Or(
                        to > symbol_factory.BitVecVal(16, 256),
                        to == symbol_factory.BitVecVal(0, 256),
                    ),
                ]
            )

            # Check whether we can also set the callee address
            try:
                constraints += [to == 0xDEADBEEFDEADBEEFDEADBEEFDEADBEEFDEADBEEF]
                solver.get_model(constraints)

                global_state.annotate(StateChangeCallsAnnotation(global_state, True))
            except UnsatError:
                global_state.annotate(StateChangeCallsAnnotation(global_state, False))
        except UnsatError:
            pass

    def _analyze_state(self, global_state: GlobalState) -> List[PotentialIssue]:

        if global_state.environment.active_function_name == "constructor":
            return []

        annotations = cast(
            List[StateChangeCallsAnnotation],
            list(global_state.get_annotations(StateChangeCallsAnnotation)),
        )
        op_code = global_state.get_current_instruction()["opcode"]

        if len(annotations) == 0 and op_code in STATE_READ_WRITE_LIST:
            return []

        if op_code in STATE_READ_WRITE_LIST:
            for annotation in annotations:
                annotation.state_change_states.append(global_state)

        # Record state changes following from a transfer of ether
        if op_code in CALL_LIST:
            value = global_state.mstate.stack[-3]  # type: BitVec
            if StateChangeAfterCall._balance_change(value, global_state):
                for annotation in annotations:
                    annotation.state_change_states.append(global_state)

        # Record external calls
        if op_code in CALL_LIST:
            StateChangeAfterCall._add_external_call(global_state)

        # Check for vulnerabilities
        vulnerabilities = []
        for annotation in annotations:
            if not annotation.state_change_states:
                continue
            issue = annotation.get_issue(global_state, self)
            if issue:
                vulnerabilities.append(issue)
        return vulnerabilities

    @staticmethod
    def _balance_change(value: BitVec, global_state: GlobalState) -> bool:
        if not value.symbolic:
            assert value.value is not None
            return value.value > 0

        else:
            constraints = copy(global_state.world_state.constraints)

            try:
                solver.get_model(
                    constraints + [value > symbol_factory.BitVecVal(0, 256)]
                )
                return True
            except UnsatError:
                return False


detector = StateChangeAfterCall()



"""This module contains representation classes for Solidity files, contracts
and source mappings."""
from typing import Dict, Set
import logging

import mythril.laser.ethereum.util as helper
from mythril.ethereum.evmcontract import EVMContract
from mythril.ethereum.util import get_solc_json
from mythril.exceptions import NoContractFoundError

log = logging.getLogger(__name__)


class SolcAST:
    def __init__(self, ast):
        self.ast = ast

    @property
    def node_type(self):
        if "nodeType" in self.ast:
            return self.ast["nodeType"]
        if "name" in self.ast:
            return self.ast["name"]
        assert False, "Unknown AST type has been fed to SolcAST"

    @property
    def abs_path(self):
        if "absolutePath" in self.ast:
            return self.ast["absolutePath"]
        else:
            return None

    @property
    def nodes(self):
        if "nodes" in self.ast:
            return self.ast["nodes"]
        if "children" in self.ast:
            return self.ast["children"]
        assert False, "Unknown AST type has been fed to SolcAST"

    def __next__(self):
        yield self.ast.__next__()

    def __getitem__(self, item):
        return self.ast[item]


class SolcSource:
    def __init__(self, source):
        self.source = source

    @property
    def ast(self):
        if "ast" in self.source:
            return SolcAST(self.source["ast"])
        if "legacyAST" in self.source:
            return SolcAST(self.source["legacyAST"])
        assert False, "Unknown source type has been fed to SolcSource"

    @property
    def id(self):
        return self.source["id"]

    @property
    def name(self):
        return self.source["name"]

    @property
    def contents(self):
        return self.source["contents"]


class SourceMapping:
    def __init__(self, solidity_file_idx, offset, length, lineno, mapping):
        """Representation of a source mapping for a Solidity file."""

        self.solidity_file_idx = solidity_file_idx
        self.offset = offset
        self.length = length
        self.lineno = lineno
        self.solc_mapping = mapping


class SolidityFile:
    """Representation of a file containing Solidity code."""

    def __init__(self, filename: str, data: str, full_contract_src_maps: Set[str]):
        """
        Metadata class containing data regarding a specific solidity file
        :param filename: The filename of the solidity file
        :param data: The code of the solidity file
        :param full_contract_src_maps: The set of contract source mappings of all the contracts in the file
        """
        self.filename = filename
        self.data = data
        self.full_contract_src_maps = full_contract_src_maps


class SourceCodeInfo:
    def __init__(self, filename, lineno, code, mapping):
        """Metadata class containing a code reference for a specific file."""

        self.filename = filename
        self.lineno = lineno
        self.code = code
        self.solc_mapping = mapping


def get_contracts_from_file(input_file, solc_settings_json=None, solc_binary="solc"):
    """

    :param input_file:
    :param solc_settings_json:
    :param solc_binary:
    """
    data = get_solc_json(
        input_file, solc_settings_json=solc_settings_json, solc_binary=solc_binary
    )

    try:
        contract_names = data["contracts"][input_file].keys()
    except KeyError:
        raise NoContractFoundError

    for contract_name in contract_names:
        if len(
            data["contracts"][input_file][contract_name]["evm"]["deployedBytecode"][
                "object"
            ]
        ):
            yield SolidityContract(
                input_file=input_file,
                name=contract_name,
                solc_settings_json=solc_settings_json,
                solc_binary=solc_binary,
            )


def get_contracts_from_foundry(input_file, foundry_json):
    """

    :param input_file:
    :param solc_settings_json:
    :param solc_binary:
    """

    try:
        contract_names = foundry_json["contracts"][input_file].keys()
    except KeyError:
        raise NoContractFoundError

    for contract_name in contract_names:
        if len(
            foundry_json["contracts"][input_file][contract_name]["evm"][
                "deployedBytecode"
            ]["object"]
        ):

            yield SolidityContract(
                input_file=input_file,
                name=contract_name,
                solc_settings_json=None,
                solc_binary=None,
                solc_data=foundry_json,
            )


class SolidityContract(EVMContract):
    """Representation of a Solidity contract."""

    def __init__(
        self,
        input_file,
        name=None,
        solc_settings_json=None,
        solc_binary="solc",
        solc_data=None,
    ):

        if solc_data is None:
            data = get_solc_json(
                input_file,
                solc_settings_json=solc_settings_json,
                solc_binary=solc_binary,
            )
        else:
            data = solc_data

        self.solc_indices = self.get_solc_indices(input_file, data)
        self.solc_json = data
        self.input_file = input_file
        has_contract = False

        # If a contract name has been specified, find the bytecode of that specific contract
        srcmap_constructor = []
        srcmap = []
        if name:
            contract = data["contracts"][input_file][name]
            if len(contract["evm"]["deployedBytecode"]["object"]):
                code = contract["evm"]["deployedBytecode"]["object"]
                creation_code = contract["evm"]["bytecode"]["object"]
                srcmap = contract["evm"]["deployedBytecode"]["sourceMap"].split(";")
                srcmap_constructor = contract["evm"]["bytecode"]["sourceMap"].split(";")
                has_contract = True

        # If no contract name is specified, get the last bytecode entry for the input file

        else:
            for contract_name, contract in sorted(
                data["contracts"][input_file].items()
            ):
                if len(contract["evm"]["deployedBytecode"]["object"]):
                    name = contract_name
                    code = contract["evm"]["deployedBytecode"]["object"]
                    creation_code = contract["evm"]["bytecode"]["object"]
                    srcmap = contract["evm"]["deployedBytecode"]["sourceMap"].split(";")
                    srcmap_constructor = contract["evm"]["bytecode"]["sourceMap"].split(
                        ";"
                    )
                    has_contract = True

        if not has_contract:
            raise NoContractFoundError

        self.mappings = []

        self.constructor_mappings = []

        self._get_solc_mappings(srcmap)
        self._get_solc_mappings(srcmap_constructor, constructor=True)

        super().__init__(code, creation_code, name=name)

    @staticmethod
    def get_sources(indices_data: Dict, source_data: Dict) -> None:
        """
        Get source indices mapping. Function not needed for older solc versions.
        """

        if "generatedSources" not in source_data:
            return
        sources = source_data["generatedSources"]
        for source in sources:
            full_contract_src_maps = SolidityContract.get_full_contract_src_maps(
                SolcAST(source["ast"])
            )
            indices_data[source["id"]] = SolidityFile(
                source["name"], source["contents"], full_contract_src_maps
            )

    @staticmethod
    def get_solc_indices(input_file: str, data: Dict) -> Dict:
        """
        Returns solc file indices
        """
        indices: Dict = {}
        for contract_data in data["contracts"].values():
            for source_data in contract_data.values():
                SolidityContract.get_sources(indices, source_data["evm"]["bytecode"])
                SolidityContract.get_sources(
                    indices, source_data["evm"]["deployedBytecode"]
                )
        for source in data["sources"].values():
            source = SolcSource(source)
            full_contract_src_maps = SolidityContract.get_full_contract_src_maps(
                source.ast
            )
            if source.ast.abs_path is not None:
                abs_path = source.ast.abs_path
            else:
                abs_path = input_file

            with open(abs_path, "rb") as f:
                code = f.read()
                indices[source.id] = SolidityFile(
                    abs_path,
                    code.decode("utf-8"),
                    full_contract_src_maps,
                )
        return indices

    @staticmethod
    def get_full_contract_src_maps(ast: SolcAST) -> Set[str]:
        """
        Takes a solc AST and gets the src mappings for all the contracts defined in the top level of the ast
        :param ast: AST of the contract
        :return: The source maps
        """
        source_maps = set()
        if ast.node_type == "SourceUnit":
            for child in ast.nodes:
                if child.get("contractKind"):
                    source_maps.add(child["src"])
        elif ast.node_type == "YulBlock":
            for child in ast["statements"]:
                source_maps.add(child["src"])

        return source_maps

    def get_source_info(self, address, constructor=False):
        """

        :param address:
        :param constructor:
        :return:
        """

        disassembly = self.creation_disassembly if constructor else self.disassembly
        mappings = self.constructor_mappings if constructor else self.mappings
        index = helper.get_instruction_index(disassembly.instruction_list, address)

        if index is None or index >= len(mappings):
            # TODO: Find why this scenario happens. Possibly an external call
            return None

        file_index = mappings[index].solidity_file_idx

        if file_index == -1:
            # If issue is detected in an internal file
            return None

        solidity_file = self.solc_indices[file_index]
        filename = solidity_file.filename

        offset = mappings[index].offset
        length = mappings[index].length

        code = solidity_file.data.encode("utf-8")[offset : offset + length].decode(
            "utf-8", errors="ignore"
        )
        lineno = mappings[index].lineno
        return SourceCodeInfo(filename, lineno, code, mappings[index].solc_mapping)

    def _is_autogenerated_code(self, offset: int, length: int, file_index: int) -> bool:
        """
        Checks whether the code is autogenerated or not
        :param offset: offset of the code
        :param length: length of the code
        :param file_index: file the code corresponds to
        :return: True if the code is internally generated, else false
        """

        if file_index == -1:
            return True
        # Handle the common code src map for the entire code.
        if (
            "{}:{}:{}".format(offset, length, file_index)
            in self.solc_indices[file_index].full_contract_src_maps
        ):
            return True

        return False

    def _get_solc_mappings(self, srcmap, constructor=False):
        """

        :param srcmap:
        :param constructor:
        """
        mappings = self.constructor_mappings if constructor else self.mappings
        prev_item = ""
        for item in srcmap:
            if item == "":
                item = prev_item
            mapping = item.split(":")

            if len(mapping) > 0 and len(mapping[0]) > 0:
                offset = int(mapping[0])

            if len(mapping) > 1 and len(mapping[1]) > 0:
                length = int(mapping[1])

            if len(mapping) > 2 and len(mapping[2]) > 0:
                idx = int(mapping[2])

            if self._is_autogenerated_code(offset, length, idx):
                lineno = None
            else:
                lineno = (
                    self.solc_indices[idx]
                    .data.encode("utf-8")[0:offset]
                    .count("\n".encode("utf-8"))
                    + 1
                )
            prev_item = item
            mappings.append(SourceMapping(idx, offset, length, lineno, item))



"""This module contains the representation for an execution state's
environment."""
from typing import Dict

from z3 import ExprRef

from mythril.laser.ethereum.state.account import Account
from mythril.laser.ethereum.state.calldata import BaseCalldata
from mythril.laser.smt import symbol_factory


class Environment:
    """The environment class represents the current execution environment for
    the symbolic executor."""

    def __init__(
        self,
        active_account: Account,
        sender: ExprRef,
        calldata: BaseCalldata,
        gasprice: ExprRef,
        callvalue: ExprRef,
        origin: ExprRef,
        basefee: ExprRef,
        code=None,
        static=False,
    ) -> None:
        """

        :param active_account:
        :param sender:
        :param calldata:
        :param gasprice:
        :param callvalue:
        :param origin:
        :param code:
        :param static: Makes the environment static.
        """
        # Metadata

        self.active_account = active_account
        self.active_function_name = ""

        self.address = active_account.address

        # TODO: Add tx_2 > tx_1 then block_no(tx_2) > block_no(tx_1)
        self.block_number = symbol_factory.BitVecSym("block_number", 256)
        self.chainid = symbol_factory.BitVecSym("chain_id", 256)

        # Ib
        self.code = active_account.code if code is None else code

        self.sender = sender
        self.calldata = calldata
        self.gasprice = gasprice
        self.origin = origin
        self.callvalue = callvalue
        self.static = static
        self.basefee = basefee

    def __str__(self) -> str:
        """

        :return:
        """
        return str(self.as_dict)

    @property
    def as_dict(self) -> Dict:
        """

        :return:
        """
        return dict(
            active_account=self.active_account,
            sender=self.sender,
            calldata=self.calldata,
            gasprice=self.gasprice,
            callvalue=self.callvalue,
            origin=self.origin,
        )
